
from nipype.interfaces.base import (BaseInterface, BaseInterfaceInputSpec,
                                    traits, File, TraitedSpec, Directory)
import numpy as np
from nipype.utils.filemanip import split_filename
import os
import glob
import shutil
import nibabel as nib
from nipype.interfaces.base import isdefined
import scipy.ndimage.measurements as snm
import datetime as dt
try:
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plot
except ImportError:
    pass
from nipype.interfaces import fsl
import pydicom
import math
import subprocess as sp


class MotionMatCalculationInputSpec(BaseInterfaceInputSpec):

    reg_mat = File(exists=True, desc='Registration matrix')
    qform_mat = File(exists=True, desc='Qform matrix')
    dummy_input = Directory(desc='Dummy input in order to make the reference '
                            'motion mat pipeline work')
    align_mats = Directory(exists=True, desc='Directory with intra-scan '
                           'alignment matrices', default=None)
    reference = traits.Bool(desc='If True, the pipeline will save just an '
                            'identity matrix (motion mats for reference scan)',
                            default=False)


class MotionMatCalculationOutputSpec(TraitedSpec):

    motion_mats = Directory(exists=True, desc='Directory with resulting motion'
                            ' matrices')


class MotionMatCalculation(BaseInterface):

    input_spec = MotionMatCalculationInputSpec
    output_spec = MotionMatCalculationOutputSpec

    def _run_interface(self, runtime):

        reference = self.inputs.reference
        dummy = self.inputs.dummy_input
        if reference:
            np.savetxt('reference_motion_mat.mat', np.eye(4))
            np.savetxt('reference_motion_mat_inv.mat', np.eye(4))
            out_name = 'ref_motion_mats'
            mm = glob.glob('*motion_mat*.mat')
        else:
            reg_mat = np.loadtxt(self.inputs.reg_mat)
            qform_mat = np.loadtxt(self.inputs.qform_mat)
            _, out_name, _ = split_filename(self.inputs.reg_mat)
            if self.inputs.align_mats:
                list_mats = sorted(glob.glob(self.inputs.align_mats + '/MAT*'))
                if not list_mats:
                    list_mats = sorted(glob.glob(
                        self.inputs.align_mats + '/*.mat'))
                    if not list_mats:
                        raise Exception(
                            'Folder {} is empty!'.format(
                                self.inputs.align_mats))
                for mat in list_mats:
                    m = np.loadtxt(mat)
                    concat = np.dot(reg_mat, m)
                    self.gen_motion_mat(concat, qform_mat, mat.split('.')[0])
                mat_path, _, _ = split_filename(mat)
                mm = glob.glob(mat_path + '/*motion_mat*.mat')
            else:
                concat = reg_mat[:]
                self.gen_motion_mat(concat, qform_mat, out_name)
                mm = glob.glob('*motion_mat*.mat')
        os.mkdir(out_name)

        for f in mm:
            shutil.move(f, out_name)

        return runtime

    def gen_motion_mat(self, concat, qform, out_name):

        concat_inv = np.linalg.inv(concat)
        concat_inv_qform = np.dot(qform, concat_inv)
        concat_inv_qform_inv = np.linalg.inv(concat_inv_qform)
        np.savetxt('{0}_motion_mat_inv.mat'.format(out_name),
                   concat_inv_qform_inv)
        np.savetxt('{0}_motion_mat.mat'.format(out_name),
                   concat_inv_qform)

    def _list_outputs(self):
        outputs = self._outputs().get()

        if self.inputs.reference:
            out_name = 'ref_motion_mats'
        else:
            _, out_name, _ = split_filename(self.inputs.reg_mat)

        outputs["motion_mats"] = os.path.abspath(out_name)

        return outputs


class MergeListMotionMatInputSpec(BaseInterfaceInputSpec):

    file_list = traits.List(mandatory=True, desc='List of files to save into '
                            'a new directory')


class MergeListMotionMatOutputSpec(TraitedSpec):

    out_dir = Directory(desc='Output directory with all the files provided as '
                        'input.')


class MergeListMotionMat(BaseInterface):
    """I created this function just to save all the matrices that are
    prodocued by MCFLIRT (whose output is a list) into a single directory.
    """
    input_spec = MergeListMotionMatInputSpec
    output_spec = MergeListMotionMatOutputSpec

    def _run_interface(self, runtime):

        file_list = self.inputs.file_list
        pth, _, _ = split_filename(file_list[0])
        if os.path.isdir(pth + '/motion_mats') is False:
            os.mkdir(pth + '/motion_mats')
        for f in file_list:
            shutil.copy(f, pth + '/motion_mats')

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()

        file_list = self.inputs.file_list
        pth, _, _ = split_filename(file_list[0])
        outputs["out_dir"] = pth + '/motion_mats'

        return outputs


class PrepareDWIInputSpec(BaseInterfaceInputSpec):

    pe_dir = traits.Str(mandatory=True, desc='Phase encoding direction, i.e. '
                        'if ROW or COL.')
    ped_polarity = traits.Float(mandatory=True, desc='phase encoding direction '
                                'polarity, i.e. if + or - ROW for example.')
    dwi = File(exists=True, desc='First diffusion image. Can '
               'be with multiple directions or b0.')
    dwi1 = File(exists=True, desc='Second diffusion image. Can'
                ' be with multiple directions or b0.')
    topup = traits.Bool(desc='Specify whether the PrepareDWI output will be'
                        'used for TOPUP distortion correction.')


class PrepareDWIOutputSpec(TraitedSpec):

    pe = traits.Str(desc='Phase encoding direction of "dwi".')
    main = File(desc='4D dwi scan for eddy or simply dwi if both dwi and dwi1 '
                'are b0s, i.e. when PrepareDWI is used before TOPUP.')
    secondary = File(desc='3D dwi scan for distortion correction in eddy or '
                     'simply dwi1 if both dwi and dwi1 are b0s, i.e. when '
                     'PrepareDWI is used before TOPUP.')
    pe_1 = traits.Str(desc='Phase encoding direction "dwi1".')


class PrepareDWI(BaseInterface):

    input_spec = PrepareDWIInputSpec
    output_spec = PrepareDWIOutputSpec

    def _run_interface(self, runtime):

        self.dict_output = {}
        pe_dir = self.inputs.pe_dir
        ped_polarity = self.inputs.ped_polarity
        topup = self.inputs.topup
        if isdefined(self.inputs.dwi) and isdefined(self.inputs.dwi1):
            dwi = nib.load(self.inputs.dwi)
            dwi = dwi.get_data()
            dwi1 = nib.load(self.inputs.dwi1)
            dwi1 = dwi1.get_data()
            if len(dwi.shape) == 4 and len(dwi1.shape) == 3:
                self.dict_output['main'] = self.inputs.dwi
                self.dict_output['secondary'] = self.inputs.dwi1
            elif len(dwi.shape) == 3 and len(dwi1.shape) == 4 and not topup:
                self.dict_output['main'] = self.inputs.dwi1
                self.dict_output['secondary'] = self.inputs.dwi
            elif len(dwi.shape) == 3 and len(dwi1.shape) == 3:
                self.dict_output['main'] = self.inputs.dwi
                self.dict_output['secondary'] = self.inputs.dwi1
            elif topup and len(dwi1.shape) == 4:
                ref = nib.load(self.inputs.dwi1)
                dwi1_b0 = dwi1[:, :, :, 0]
                im2save = nib.Nifti1Image(dwi1_b0, affine=ref.affine)
                nib.save(im2save, 'b0.nii.gz')
                self.dict_output['main'] = self.inputs.dwi
                self.dict_output['secondary'] = os.getcwd() + '/b0.nii.gz'

        if np.sign(ped_polarity) == 1:
            if pe_dir == 'ROW':
                self.dict_output['pe'] = 'RL'
            elif pe_dir == 'COL':
                self.dict_output['pe'] = 'AP'
        elif np.sign(ped_polarity) == -1:
            if pe_dir == 'ROW':
                self.dict_output['pe'] = 'LR'
            elif pe_dir == 'COL':
                self.dict_output['pe'] = 'PA'
        self.dict_output['pe_1'] = self.dict_output['pe'][::-1]

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["pe"] = self.dict_output['pe']
        if isdefined(self.inputs.dwi) and isdefined(self.inputs.dwi1):
            outputs["main"] = self.dict_output['main']
            outputs["secondary"] = self.dict_output['secondary']
        if isdefined(self.dict_output['pe_1']):
            outputs["pe_1"] = self.dict_output['pe_1']

        return outputs


class CheckDwiNamesInputSpec(BaseInterfaceInputSpec):

    dicom_dwi = Directory(desc='First diffusion image (in DICOM format), '
                          'provided as input.')
    dicom_dwi1 = Directory(desc='Second diffusion image (in DICOM format), '
                           'provided as input.')
    nifti_dwi = File(desc='Main dwi image as detected by PrepareDWI.')


class CheckDwiNamesOutputSpec(TraitedSpec):

    main = Directory()


class CheckDwiNames(BaseInterface):
    """
    Since in PrepareDWI I have to use nifti files to get the data shape while
    in dwipreproc I need to provide DICOM, I created this interface to see
    which of the dwi DICOM files provided as input are 4D and 3D files and
    assign them correctly in dwipreproc node.
    """
    input_spec = CheckDwiNamesInputSpec
    output_spec = CheckDwiNamesOutputSpec

    def _run_interface(self, runtime):

        dwi = self.inputs.dicom_dwi
        dwi1 = self.inputs.dicom_dwi1
        nifti = self.inputs.nifti_dwi
        self.dict_output = {}

        if dwi.split('/')[-1] in nifti:
            self.dict_output['main'] = self.inputs.dicom_dwi
        elif dwi1.split('/')[-1] in nifti:
            self.dict_output['main'] = self.inputs.dicom_dwi1

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["main"] = self.dict_output['main']

        return outputs


class GenTopupConfigFilesInputSpec(BaseInterfaceInputSpec):

    ped = traits.Str(desc='phase encoding direction for the main image')


class GenTopupConfigFilesOutputSpec(TraitedSpec):

    config_file = File(exists=True, desc='configuration file for topup')
    apply_topup_config = File(
        exists=True, desc='configuration file for apply_topup')


class GenTopupConfigFiles(BaseInterface):

    input_spec = GenTopupConfigFilesInputSpec
    output_spec = GenTopupConfigFilesOutputSpec

    def _run_interface(self, runtime):

        ped = self.inputs.ped

        if ped == 'RL':
            with open('config_file.txt', 'w') as f:
                f.write('-1 0 0 1 \n1 0 0 1')
            f.close()
            with open('apply_topup_config_file.txt', 'w') as f:
                f.write('-1 0 0 1')
            f.close()
        elif ped == 'LR':
            with open('config_file.txt', 'w') as f:
                f.write('1 0 0 1 \n-1 0 0 1')
            f.close()
            with open('apply_topup_config_file.txt', 'w') as f:
                f.write('1 0 0 1')
            f.close()
        elif ped == 'AP':
            with open('config_file.txt', 'w') as f:
                f.write('0 -1 0 1 \n0 1 0 1')
            f.close()
            with open('apply_topup_config_file.txt', 'w') as f:
                f.write('0 -1 0 1')
            f.close()
        elif ped == 'AP':
            with open('config_file.txt', 'w') as f:
                f.write('0 1 0 1 \n0 -1 0 1')
            f.close()
            with open('apply_topup_config_file.txt', 'w') as f:
                f.write('0 1 0 1')
            f.close()

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["config_file"] = os.path.join(
            os.getcwd(), 'config_file.txt')
        outputs["apply_topup_config"] = os.path.join(
            os.getcwd(), 'apply_topup_config_file.txt')

        return outputs


class AffineMatrixGenerationInputSpec(BaseInterfaceInputSpec):

    motion_parameters = File(exists=True, desc='File with all the motion '
                             'paramters calculated by eddy.')
    reference_image = File(exists=True, desc='Reference used to calculate the '
                           'motion (usually is the first volume).')


class AffineMatrixGenerationOutputSpec(TraitedSpec):

    affine_matrices = Directory(exists=True, desc='Directory containing all '
                                'affine matrices calculated by the interface.')


class AffineMatrixGeneration(BaseInterface):
    """
    Given the six motion parameters calculated by eddy, this function creates
    the 4x4 affine transformations that align each diffusion direction to the
    first volume (usually a b0).
    """
    input_spec = AffineMatrixGenerationInputSpec
    output_spec = AffineMatrixGenerationOutputSpec

    def _run_interface(self, runtime):

        _, out_name, _ = split_filename(self.inputs.motion_parameters)
        motion_par = np.loadtxt(self.inputs.motion_parameters)
        motion_par = motion_par[:, :6]
        ref = nib.load(self.inputs.reference_image)
        ref_data = ref.get_data()
        # centre of mass
        if len(ref_data.shape) == 4:
            com = np.asarray(snm.center_of_mass(ref_data[:, :, :, 0]))
        else:
            com = np.asarray(snm.center_of_mass(ref_data))

        hdr = ref.header
        resolution = list(hdr.get_zooms()[:3])

        for i in range(len(motion_par)):
            mat = self.create_affine_mat(motion_par[i, :], resolution * com)
            np.savetxt(
                'affine_mat_{}.mat'.format(str(i).zfill(4)), mat, fmt='%f')

        affines = glob.glob('affine_mat*.mat')
        os.mkdir(out_name)
        for f in affines:
            shutil.move(f, out_name)

        return runtime

    def create_affine_mat(self, mp, cog):

        T = np.eye(4)

        T[0, -1] = cog[0]
        T[1, -1] = cog[1]
        T[2, -1] = cog[2]

        T_1 = np.linalg.inv(T)

        tx = mp[0]
        ty = mp[1]
        tz = mp[2]
        rx = mp[3]
        ry = mp[4]
        rz = mp[5]

        Rx = np.eye(3)
        Rx[1, 1] = np.cos(rx)
        Rx[1, 2] = np.sin(rx)
        Rx[2, 1] = -np.sin(rx)
        Rx[2, 2] = np.cos(rx)
        Ry = np.eye(3)
        Ry[0, 0] = np.cos(ry)
        Ry[0, 2] = -np.sin(ry)
        Ry[2, 0] = np.sin(ry)
        Ry[2, 2] = np.cos(ry)
        Rz = np.eye(3)
        Rz[0, 0] = np.cos(rz)
        Rz[0, 1] = np.sin(rz)
        Rz[1, 0] = -np.sin(rz)
        Rz[1, 1] = np.cos(rz)
        R_3x3 = np.dot(np.dot(Rx, Ry), Rz)

        m = np.eye(4)
        m[:3, :3] = R_3x3
        m[0, 3] = tx
        m[1, 3] = ty
        m[2, 3] = tz

        new_orig = np.dot(T, np.dot(m, T_1))[:, -1]

        m[0, 3] = new_orig[0]
        m[1, 3] = new_orig[1]
        m[2, 3] = new_orig[2]

        return m

    def _list_outputs(self):
        outputs = self._outputs().get()

        _, out_name, _ = split_filename(self.inputs.motion_parameters)

        outputs["affine_matrices"] = os.path.abspath(out_name)

        return outputs


class MeanDisplacementCalculationInputSpec(BaseInterfaceInputSpec):

    motion_mats = traits.List(desc='List of motion mats.')
    trs = traits.List(desc='List of repetition times.')
    start_times = traits.List(desc='List of start times.')
    real_durations = traits.List(desc='List of real durations.')
    reference = File(desc='Reference image.')
    input_names = traits.List(desc='List with the name of the inputs provided'
                              'by the user. This is used for the severe motion'
                              'report in order to provide the real names of '
                              'possibly corrupted images.')


class MeanDisplacementCalculationOutputSpec(TraitedSpec):

    mean_displacement = File(exists=True, desc='mean displacement between each'
                             ' scan/volume and the reference.')
    mean_displacement_rc = File(exists=True, desc='mean displacement values '
                                'used in the generate the plot. Essentially it'
                                ' contains one mean displacement value per '
                                'millisecond. If, for example, a study lasts '
                                'for 60 minutes, this file will contain '
                                '3600000 values. I created this file to be '
                                'able to plot the entire scan time, including '
                                'both real scan time and MR idling time.')
    mean_displacement_consecutive = File(exists=True, desc='mean displacement '
                                         'between each pair of consecutive '
                                         'scans/volumes.')
    start_times = File(exists=True, desc='start times for each scan/volume.')
    motion_parameters_rc = File(
        exists=True, desc='Same as mean_displacement_rc but for motion '
        'parameters.')
    motion_parameters = File(exists=True, desc='6 motion parameters (3 '
                             'rotation and 3 translation) per scan/volume.')
    offset_indexes = File(exists=True, desc='indexes where the '
                          'mean_displacement_rc values reflect MR idling '
                          'times. Used in the plot.')
    mats4average = File(exists=True, desc='location of all the motion matrices'
                        ' used to calculate the mean displacement. This will '
                        'be used to create an average motion mat per detected '
                        'frame.')
    corrupted_volumes = File(exists=True, desc='report of any unusually severe'
                             ' motion detected.')


class MeanDisplacementCalculation(BaseInterface):

    input_spec = MeanDisplacementCalculationInputSpec
    output_spec = MeanDisplacementCalculationOutputSpec

    def _run_interface(self, runtime):

        list_inputs = list(zip(
            self.inputs.motion_mats, self.inputs.start_times,
            self.inputs.real_durations, self.inputs.trs,
            self.inputs.input_names))
        ref = nib.load(self.inputs.reference)
        ref_data = ref.get_data()
        # centre of gravity
        ref_cog = np.asarray(snm.center_of_mass(ref_data))
        list_inputs = sorted(list_inputs, key=lambda k: k[1])
        study_start_time = list_inputs[0][1]
        list_inputs = [
            (x[0], (dt.datetime.strptime(x[1], '%H%M%S.%f') -
                    dt.datetime.strptime(list_inputs[0][1], '%H%M%S.%f'))
             .total_seconds(), x[2], x[3], x[4]) for x in list_inputs]
        study_len = int((list_inputs[-1][1] + float(list_inputs[-1][2])) *
                        1000)
        mean_displacement_rc = np.zeros(study_len) - 1
        motion_par_rc = np.zeros((6, study_len)) - 1
        mean_displacement = []
        motion_par = []
        idt_mat = np.eye(4)
        all_mats = []
        all_mats4average = []
        start_times = []
        volume_names = []
        corrupted_volume_names = [
            'No volume showed rotation greater than 8 degrees and/or '
            'translation greater than 20mm respect to the reference.\nHowever,'
            ' have a look at the mean displacement plot to see whether there'
            ' are scans with very different mean displacement with respect '
            'to the others.\nIn that case please check the registration of '
            'that particular scan.']
        for f in list_inputs:
            mats = sorted(glob.glob(f[0] + '/*inv.mat'))
            mats4averge = sorted(glob.glob(f[0] + '/*mat.mat'))
            all_mats = all_mats + mats
            all_mats4average = all_mats4average + mats4averge
            start_scan = f[1]
            tr = f[3]
            if len(mats) > 1:  # for 4D files
                for i, mat in enumerate(mats):
                    volume_names.append(f[-1] + '_vol_{}'
                                        .format(str(i + 1).zfill(4)))
                    start_times.append((
                        dt.datetime.strptime(
                            study_start_time, '%H%M%S.%f') +
                        dt.timedelta(seconds=start_scan)).strftime(
                            '%H%M%S.%f'))
                    end_scan = start_scan + tr
                    m = np.loadtxt(mat)
                    md = self.rmsdiff(ref_cog, m, idt_mat)
                    mean_displacement_rc[
                        int(start_scan * 1000):int(end_scan * 1000)] = md
                    mean_displacement.append(md)
                    mp = self.avscale(m, ref_cog)
                    motion_par.append(mp)
                    duration = int(end_scan * 1000) - int(start_scan * 1000)
                    motion_par_rc[:, int(start_scan * 1000):
                                  int(end_scan * 1000)] = np.array(
                                      [mp, ] * duration).T
                    start_scan = end_scan
            elif len(mats) == 1:  # for 3D files
                volume_names.append(f[-1])
                start_times.append((
                    dt.datetime.strptime(study_start_time,
                                         '%H%M%S.%f') +
                    dt.timedelta(seconds=start_scan)).strftime('%H%M%S.%f'))
                end_scan = start_scan + float(f[2])
                m = np.loadtxt(mats[0])
                md = self.rmsdiff(ref_cog, m, idt_mat)
                mean_displacement_rc[
                    int(start_scan * 1000):int(end_scan * 1000)] = md
                mean_displacement.append(md)
                mp = self.avscale(m, ref_cog)
                motion_par.append(mp)
                duration = int(end_scan * 1000) - int(start_scan * 1000)
                motion_par_rc[:, int(start_scan * 1000):
                              int(end_scan * 1000)] = np.array(
                                  [mp, ] * duration).T
        start_times.append((
            dt.datetime.strptime(study_start_time, '%H%M%S.%f') +
            dt.timedelta(seconds=end_scan)).strftime('%H%M%S.%f'))
        mean_displacement_consecutive = []
        for i in range(len(all_mats) - 1):
            m1 = np.loadtxt(all_mats[i])
            m2 = np.loadtxt(all_mats[i + 1])
            md_consecutive = self.rmsdiff(ref_cog, m1, m2)
            mean_displacement_consecutive.append(md_consecutive)

        corrupted_volumes = self.check_max_motion(motion_par)
        if corrupted_volumes:
            corrupted_volume_names = [
                'The following volumes showed an unusual severe motion (i.e. '
                'rotation greater than 8 degrees and/or translation greater '
                'than 20mm). \nThis is suspiciuos and can be potentially due '
                'to errors in the registration process. Please check the '
                'specified images before moving farward.\n']
            corrupted_volume_names = (
                corrupted_volume_names + [volume_names[x]
                                          for x in corrupted_volumes])
        offset_indexes = np.where(mean_displacement_rc == -1)
        for i in range(len(mean_displacement_rc)):
            if (mean_displacement_rc[i] == -1 and
                    mean_displacement_rc[i - 1] != -1):
                mean_displacement_rc[i] = mean_displacement_rc[i - 1]
        for i in range(motion_par_rc.shape[1]):
            if (motion_par_rc[0, i] == -1 and motion_par_rc[0, i - 1] != -1):
                motion_par_rc[:, i] = motion_par_rc[:, i - 1]

        to_save = [mean_displacement, mean_displacement_consecutive,
                   mean_displacement_rc, motion_par_rc, start_times,
                   offset_indexes, all_mats4average, motion_par,
                   corrupted_volume_names]
        to_save_name = ['mean_displacement', 'mean_displacement_consecutive',
                        'mean_displacement_rc', 'motion_par_rc', 'start_times',
                        'offset_indexes', 'mats4average', 'motion_par',
                        'severe_motion_detection_report']
        for i in range(len(to_save)):
            np.savetxt(to_save_name[i] + '.txt', np.asarray(to_save[i]),
                       fmt='%s')

        return runtime

    def rmsdiff(self, cog, T1, T2):
        """Python implementation of the rmsdiff function in fsl"""
        R = 80
        M = np.dot(T2, np.linalg.inv(T1)) - np.identity(4)
        A = M[:3, :3]
        t = M[:3, 3]
        Tr = np.trace(np.dot(A.T, A))
        II = (t + np.dot(A, cog)).T
        III = t + np.dot(A, cog)
        cost = Tr * R ** 2 / 5
        rms = np.sqrt(cost + np.dot(II, III))

        return rms

    def avscale(self, mat, com, res=[1, 1, 1], moco=False):
        """Python implementation of the avscale function in fsl. However this
        works just with affine matrices from rigid body motion, i.e. it assumes
        that there is no scales or skew effect. Furthermore, if moco=True,
        it returns the rigid body motion parameters in Siemens moco series
        convetion."""
        c = np.asarray(com)
        trans_init = mat[:3, -1]
        rot_mat = mat[:3, :3]
        centre = c * res
        rot_x, rot_y, rot_z = self.rotationMatrixToEulerAngles(rot_mat)
        trans_tot = np.dot(rot_mat, centre) + trans_init - centre
        trans_x = trans_tot[0]
        trans_y = trans_tot[1]
        trans_z = trans_tot[2]
        if moco:
            rot_x_moco = -self.rad2degree(rot_y)
            rot_y_moco = self.rad2degree(rot_x)
            rot_z_moco = -self.rad2degree(rot_z)
            trans_x_moco = -trans_y
            trans_y_moco = trans_x
            trans_z_moco = -trans_z
            print([trans_x_moco, trans_y_moco, trans_z_moco, rot_x_moco,
                   rot_y_moco, rot_z_moco])
        return [rot_x, rot_y, rot_z, trans_x, trans_y, trans_z]

    def rad2degree(self, alpha_rad):
        return alpha_rad * 180 / np.pi

    def isRotationMatrix(self, R):
        Rt = np.transpose(R)
        shouldBeIdentity = np.dot(Rt, R)
        Identity = np.identity(3, dtype=R.dtype)
        n = np.linalg.norm(Identity - shouldBeIdentity)
        return n < 1e-4

    def rotationMatrixToEulerAngles(self, R):
        assert(self.isRotationMatrix(R))
        cy = math.sqrt(R[0, 0] * R[0, 0] + R[0, 1] * R[0, 1])
        singular = cy < 1e-4
        if not singular:
            cz = R[0, 0] / cy
            sz = R[0, 1] / cy
            cx = R[2, 2] / cy
            sx = R[1, 2] / cy
            sy = -R[0, 2]
            x = math.atan2(sx, cx)
            y = math.atan2(sy, cy)
            z = math.atan2(sz, cz)
        else:
            cx = R[1, 1]
            sx = -R[2, 1]
            sy = -R[0, 2]
            x = math.atan2(sx, cx)
            y = math.atan2(sy, 0.0)
            z = 0.0
        return np.array([x, y, z])

    def check_max_motion(self, motion_par):

        corrupted_vol_rot = np.where(np.abs(
            np.asarray(motion_par)[:, :3]) >= 0.14)[0]
        corrupted_vol_trans = np.where(np.abs(
            np.asarray(motion_par)[:, 3:]) >= 20)[0]
        corrupted_vol = list(set(corrupted_vol_rot.tolist() +
                                 corrupted_vol_trans.tolist()))

        return corrupted_vol

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["mean_displacement"] = os.getcwd() + '/mean_displacement.txt'
        outputs["mean_displacement_rc"] = (
            os.getcwd() + '/mean_displacement_rc.txt')
        outputs["mean_displacement_consecutive"] = (
            os.getcwd() + '/mean_displacement_consecutive.txt')
        outputs["start_times"] = os.getcwd() + '/start_times.txt'
        outputs["motion_parameters"] = os.getcwd() + '/motion_par.txt'
        outputs["motion_parameters_rc"] = os.getcwd() + '/motion_par_rc.txt'
        outputs["offset_indexes"] = os.getcwd() + '/offset_indexes.txt'
        outputs["mats4average"] = os.getcwd() + '/mats4average.txt'
        outputs["corrupted_volumes"] = (
            os.getcwd() + '/severe_motion_detection_report.txt')

        return outputs


class MotionFramingInputSpec(BaseInterfaceInputSpec):

    mean_displacement = File(exists=True)
    mean_displacement_consec = File(exists=True)
    start_times = File(exists=True)
    motion_threshold = traits.Float(desc='Everytime the mean displacement is '
                                    'greater than this value (in mm), a new '
                                    'frame will be initialised. Default 2mm',
                                    default=2)
    temporal_threshold = traits.Float(desc='If one frame temporal duration is '
                                      'shorter than this value (in sec) then '
                                      'the frame will be discarded. Default '
                                      '30sec', default=30)
    pet_start_time = traits.Str(desc='PET start time', default=None)
    pet_end_time = traits.Str(desc='PET end time', default=None)
    pet_offset = traits.Int(desc='Offset in seconds between the PET start time'
                            'and the time you want to start the static motion'
                            'correction. Default is 0.')
    pet_duration = traits.Int(desc='Time, in seconds, the static PET '
                              'reconstruction lasts. Default is from '
                              'pet_start_time+pet_offest to the pet_end_time')


class MotionFramingOutputSpec(TraitedSpec):

    frame_start_times = File(exists=True, desc='start times of each of the '
                             'detected frame in real clock time.')
    frame_vol_numbers = File(exists=True, desc='Text file with the number of '
                             'volume where the motion occurred.')
    timestamps_dir = Directory(desc='Directory with the timestamps for all'
                               ' the detected frames')


class MotionFraming(BaseInterface):

    input_spec = MotionFramingInputSpec
    output_spec = MotionFramingOutputSpec

    def _run_interface(self, runtime):

        mean_displacement = np.loadtxt(self.inputs.mean_displacement,
                                       dtype=float)
        mean_displacement_consecutive = np.loadtxt(
            self.inputs.mean_displacement_consec, dtype=float)
        th = self.inputs.motion_threshold
        start_times = np.loadtxt(self.inputs.start_times, dtype=str)
        temporal_th = self.inputs.temporal_threshold
        pet_st = self.inputs.pet_start_time
        pet_endtime = self.inputs.pet_end_time
        if not pet_st and not pet_endtime:
            pet_st = None
            pet_endtime = None
        else:
            if isdefined(self.inputs.pet_offset):
                offset = self.inputs.pet_offset
                pet_st = (dt.datetime.strptime(pet_st, '%H%M%S.%f') +
                          dt.timedelta(seconds=offset)).strftime('%H%M%S.%f')
            if (isdefined(self.inputs.pet_duration) and
                    self.inputs.pet_duration > 0):
                pet_len = self.inputs.pet_duration
                pet_endtime = ((dt.datetime.strptime(pet_st, '%H%M%S.%f') +
                                dt.timedelta(seconds=pet_len))
                               .strftime('%H%M%S.%f'))

        md_0 = mean_displacement[0]
        max_md = mean_displacement[0]
        frame_vol = [0]
        frame_st4pet = []

        scan_duration = [
            (dt.datetime.strptime(start_times[i], '%H%M%S.%f') -
             dt.datetime.strptime(start_times[i - 1], '%H%M%S.%f')
             ).total_seconds() for i in range(1, len(start_times))]

        for i, md in enumerate(mean_displacement[1:]):

            current_md = md
            if (np.abs(md_0 - current_md) > th or
                    np.abs(max_md - current_md) > th):
                duration = np.sum(scan_duration[frame_vol[-1]:i + 1])
                if duration > temporal_th:
                    if i + 1 not in frame_vol:
                        frame_vol.append(i + 1)

                    md_0 = current_md
                    max_md = current_md
                else:
                    prev_md = mean_displacement[frame_vol[-1]]
                    if (prev_md - current_md) > th * 2:
                        frame_vol.remove(frame_vol[-1])
                    elif (current_md - prev_md) > th:
                        frame_vol.remove(frame_vol[-1])
                        frame_vol.append(i)
            elif mean_displacement_consecutive[i] > th:
                duration = np.sum(scan_duration[frame_vol[-1]:i + 1])
                if duration > temporal_th:
                    if i + 1 not in frame_vol:
                        frame_vol.append(i + 1)
                    md_0 = current_md
                    max_md = current_md
            elif current_md > max_md:
                max_md = current_md
            elif current_md < md_0:
                md_0 = current_md

        duration = np.sum(scan_duration[frame_vol[-1]:i + 2])
        if duration > temporal_th:
            if (i + 2) not in frame_vol:
                frame_vol.append(i + 2)
        else:
            frame_vol.remove(frame_vol[-1])
            frame_vol.append(i + 2)

        frame_vol = sorted(frame_vol)
        frame_start_times = [start_times[x] for x in frame_vol]
        if pet_st and pet_endtime:
            frame_st4pet = [
                x for x in frame_start_times if
                (dt.datetime.strptime(x, '%H%M%S.%f') >
                 dt.datetime.strptime(pet_st, '%H%M%S.%f')
                 and dt.datetime.strptime(x, '%H%M%S.%f') <
                 dt.datetime.strptime(pet_endtime, '%H%M%S.%f'))]
            if frame_start_times[0] in frame_st4pet:
                frame_st4pet.remove(frame_start_times[0])
            if frame_start_times[-1] in frame_st4pet:
                frame_st4pet.remove(frame_start_times[-1])
            frame_st4pet.append(pet_st)
            frame_st4pet.append(pet_endtime)
            frame_st4pet = sorted(frame_st4pet)
            if ((dt.datetime.strptime(frame_st4pet[1], '%H%M%S.%f') -
                    dt.datetime.strptime(frame_st4pet[0], '%H%M%S.%f'))
                    .total_seconds() < 30):
                frame_st4pet.remove(frame_st4pet[1])
            if ((dt.datetime.strptime(frame_st4pet[-1], '%H%M%S.%f') -
                    dt.datetime.strptime(frame_st4pet[-2], '%H%M%S.%f'))
                    .total_seconds() < 30):
                frame_st4pet.remove(frame_st4pet[-2])
            frame_vol = [i for i in range(len(start_times)) for j in
                         range(len(frame_st4pet)) if start_times[i] ==
                         frame_st4pet[j]]
            if (dt.datetime.strptime(start_times[0], '%H%M%S.%f') >
                    dt.datetime.strptime(pet_st, '%H%M%S.%f')):
                frame_vol.append(0)
            else:
                vol = [i for i in range(len(start_times)) if
                       (dt.datetime.strptime(start_times[i], '%H%M%S.%f') <
                        dt.datetime.strptime(frame_st4pet[0], '%H%M%S.%f') and
                        dt.datetime.strptime(start_times[i + 1], '%H%M%S.%f') >
                        dt.datetime.strptime(frame_st4pet[0], '%H%M%S.%f'))]
                frame_vol.append(vol[0])
            if (dt.datetime.strptime(start_times[-1], '%H%M%S.%f') <
                    dt.datetime.strptime(pet_endtime, '%H%M%S.%f')):
                frame_vol.append(len(start_times) - 1)
            else:
                vol = [i for i in range(len(start_times)) if
                       (dt.datetime.strptime(start_times[i], '%H%M%S.%f') <
                        dt.datetime.strptime(frame_st4pet[-1], '%H%M%S.%f') and
                        dt.datetime.strptime(start_times[i + 1], '%H%M%S.%f') >
                        dt.datetime.strptime(frame_st4pet[-1], '%H%M%S.%f'))]
                frame_vol.append(vol[0])
            frame_vol = sorted(frame_vol)
        np.savetxt('frame_start_times.txt', np.asarray(frame_start_times),
                   fmt='%s')
        os.mkdir('timestamps')
        if frame_st4pet:
            timestamps_2save = frame_st4pet
        else:
            timestamps_2save = frame_start_times
        np.savetxt('timestamps/frame_start_times_4PET.txt',
                   np.asarray(timestamps_2save), fmt='%s')
        for i in range(len(timestamps_2save) - 1):
            with open('timestamps/timestamps_Frame{}.txt'
                      .format(str(i).zfill(3)), 'w') as f:
                f.write(timestamps_2save[i] + '\n' + timestamps_2save[i + 1])
            f.close()
        np.savetxt('frame_vol_numbers.txt', np.asarray(frame_vol), fmt='%s')

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["frame_start_times"] = os.getcwd() + '/frame_start_times.txt'
        outputs["frame_vol_numbers"] = os.getcwd() + '/frame_vol_numbers.txt'
        outputs["timestamps_dir"] = os.getcwd() + '/timestamps'

        return outputs


class PlotMeanDisplacementRCInputSpec(BaseInterfaceInputSpec):

    mean_disp_rc = File(exists=True, desc='Text file containing the mean '
                        'displacement real clock.')
    motion_par_rc = File(exists=True, desc='Text file containing the motion '
                         'parameters real clock.')
    frame_start_times = File(exists=True, desc='Frame start times as detected'
                             'by the motion framing pipeline')
    false_indexes = File(exists=True, desc='Time indexes were the scanner was '
                         'idling, i.e. there is no motion information.')
    framing = traits.Bool(desc='If true, the frame start times will be plotted'
                          'in the final image.')


class PlotMeanDisplacementRCOutputSpec(TraitedSpec):

    mean_disp_plot = File(exists=True, desc='Mean displacement plot.')
    rot_plot = File(exists=True, desc='Rotation parameters plot.')
    trans_plot = File(exists=True, desc='Translation parameters plot.')


class PlotMeanDisplacementRC(BaseInterface):

    input_spec = PlotMeanDisplacementRCInputSpec
    output_spec = PlotMeanDisplacementRCOutputSpec

    def _run_interface(self, runtime):

        mean_disp_rc = np.loadtxt(self.inputs.mean_disp_rc)
        false_indexes = np.loadtxt(self.inputs.false_indexes, dtype=int)

        if isdefined(self.inputs.motion_par_rc):
            motion_par_rc = np.loadtxt(self.inputs.motion_par_rc)
            plot_mp = True
        else:
            plot_mp = False
        plot_offset = True
        dates = np.arange(0, len(mean_disp_rc), 1)
        indxs = np.zeros(len(mean_disp_rc), int) + 1
        indxs[false_indexes] = 0
        start_true_period = [x for x in range(1, len(indxs)) if indxs[x] == 1
                             and indxs[x - 1] == 0]
        end_true_period = [x for x in range(len(indxs) - 1) if indxs[x] == 0 and
                           indxs[x - 1] == 1]
        start_true_period.append(1)
        end_true_period.append(len(dates))
        start_true_period = sorted(start_true_period)
        end_true_period = sorted(end_true_period)
        if len(start_true_period) == len(end_true_period) - 1:
            end_true_period.remove(end_true_period[-1])
        elif len(start_true_period) != len(end_true_period):
            print ('Something went wrong in the identification of the MR '
                   'idling time. It will not be plotted.')
            plot_offset = False

        self.gen_plot(dates, mean_disp_rc, plot_offset, start_true_period,
                      end_true_period)
        if plot_mp:
            for i in range(2):
                mp = motion_par_rc[i * 3:(i + 1) * 3, :]
                self.gen_plot(
                    dates, mp, plot_offset, start_true_period, end_true_period,
                    plot_mp=plot_mp, mp_ind=i)

        return runtime

    def gen_plot(self, dates, to_plot, plot_offset, start_true_period,
                 end_true_period, plot_mp=False, mp_ind=None):

        frame_start_times = np.loadtxt(self.inputs.frame_start_times)
        framing = self.inputs.framing
        font = {'weight': 'bold', 'size': 30}
        matplotlib.rc('font', **font)
        fig, ax = plot.subplots()
        fig.set_size_inches(21, 9)
        ax.set_xlim(0, dates[-1])
        ax.set_ylim(25, 60)
        if plot_mp:
            col = ['b', 'g', 'r']
        if plot_offset:
            for i in range(0, len(start_true_period)):
                if plot_mp:
                    for ii in range(3):
                        ax.plot(
                            dates[start_true_period[i] - 1:end_true_period[i] + 1],
                            to_plot[ii, start_true_period[i] - 1:
                                    end_true_period[i] + 1],
                            c=col[ii], linewidth=2)
                else:
                    ax.plot(dates[start_true_period[i] - 1:end_true_period[i] + 1],
                            to_plot[start_true_period[i] - 1:
                                    end_true_period[i] + 1], c='b', linewidth=2)
            for i in range(0, len(end_true_period) - 1):
                if plot_mp:
                    for ii in range(3):
                        ax.plot(
                            dates[end_true_period[i] - 1:
                                  start_true_period[i + 1] + 1],
                            to_plot[ii, end_true_period[i] - 1:
                                    start_true_period[i + 1] + 1],
                            c=col[ii], linewidth=2, ls='--', dashes=(2, 3))
                else:
                    ax.plot(
                        dates[end_true_period[i] - 1:start_true_period[i + 1] + 1],
                        to_plot[end_true_period[i] - 1:start_true_period[i + 1] + 1],
                        c='b', linewidth=2, ls='--', dashes=(2, 3))

        if framing:
            cl = 'yellow'
            for i in range(len(frame_start_times[:-1])):

                tt = (
                    (dt.datetime.strptime(str(frame_start_times[i]),
                                          '%H%M%S.%f') -
                     dt.datetime.strptime(str(frame_start_times[0]),
                                          '%H%M%S.%f'))
                    .total_seconds() * 1000)
                if tt >= len(dates):
                    tt = len(dates) - 1
                plot.axvline(dates[int(tt)], c='b', alpha=0.3, ls='--')

                tt1 = ((dt.datetime.strptime(str(frame_start_times[i + 1]),
                                             '%H%M%S.%f') -
                       dt.datetime.strptime(str(frame_start_times[0]),
                                            '%H%M%S.%f'))
                       .total_seconds() * 1000)
                if tt1 >= len(dates):
                    tt1 = len(dates) - 1
                plot.axvspan(dates[int(tt)], dates[int(tt1)], facecolor=cl,
                             alpha=0.4, linewidth=0)

                if i % 2 == 0:
                    cl = 'w'
                else:
                    cl = 'yellow'

        indx = np.arange(0, len(dates), 300000)
        my_thick = [str(i) for i in np.arange(0, len(dates) / 60000, 5,
                                              dtype=int)]
        plot.xticks(dates[indx], my_thick)
#         ax.set_yscale('log')
#         ax.set_yticks([10, 30, 50])
#         ax.get_yaxis().set_major_formatter(matplotlib.ticker.ScalarFormatter())
        my_y_ticks = [25, 30, 35, 40, 45, 50, 55]
#         y = [0, 1, 2, 3, 4, 5, 10, 30, 35, 40, 45, 50, 55, 60]
        plot.yticks(my_y_ticks, my_y_ticks)
        plot.xlabel('Time [min]', fontsize=25)
        if mp_ind == 0:
            ax.set_ylim(np.min(to_plot) - 0.1, np.max(to_plot) + 0.1)
            plot.legend(['Rotation X', 'Rotation Y', 'Rotation Z'], loc=0)
            plot.ylabel('Rotation [rad]', fontsize=25)
            plot.savefig('Rotation_real_clock.png')
        elif mp_ind == 1:
            ax.set_ylim(np.min(to_plot) - 0.5, np.max(to_plot) + 0.5)
            plot.legend(
                ['Translation X', 'Translation Y', 'Translation Z'], loc=0)
            plot.ylabel('Translation [mm]', fontsize=25)
            plot.savefig('Translation_real_clock.png')
        else:
            plot.ylabel('Mean displacement [mm]', fontsize=25)
            plot.savefig('mean_displacement_real_clock.png')
        plot.close()

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["mean_disp_plot"] = (
            os.getcwd() + '/mean_displacement_real_clock.png')
        outputs["rot_plot"] = (
            os.getcwd() + '/Rotation_real_clock.png')
        outputs["trans_plot"] = (
            os.getcwd() + '/Translation_real_clock.png')

        return outputs


class AffineMatAveragingInputSpec(BaseInterfaceInputSpec):

    frame_vol_numbers = File(exists=True)
    all_mats4average = File(exists=True)


class AffineMatAveragingOutputSpec(TraitedSpec):

    average_mats = Directory(exists=True, desc='directory with all the average'
                             ' transformation matrices for each detected '
                             'frame.')


class AffineMatAveraging(BaseInterface):

    input_spec = AffineMatAveragingInputSpec
    output_spec = AffineMatAveragingOutputSpec

    def _run_interface(self, runtime):

        frame_vol = np.loadtxt(self.inputs.frame_vol_numbers, dtype=int)
        all_mats = np.loadtxt(self.inputs.all_mats4average, dtype=str)
        idt = np.eye(4)

        for v in range(len(frame_vol) - 1):

            v1 = frame_vol[v]
            v2 = frame_vol[v + 1]
            mat_tot = np.zeros((4, 4, (v2 - v1)))
            n_vol = 0

            for j, m in enumerate(all_mats[v1:v2]):
                mat = np.loadtxt(m)
                if (mat == idt).all():
                    mat_tot[:, :, j] = np.zeros((4, 4))
                else:
                    mat_tot[:, :, j] = mat
                    n_vol += 1
            if n_vol > 0:
                average_mat = np.sum(mat_tot, axis=2) / n_vol
            else:
                average_mat = idt

            np.savetxt(
                'average_matrix_vol_{0}-{1}.txt'
                .format(str(v1).zfill(4), str(v2).zfill(4)), average_mat)

        if os.path.isdir('frame_mean_transformation_mats') is False:
            os.mkdir('frame_mean_transformation_mats')

        mats = glob.glob('average_matrix_vol*')
        for m in mats:
            shutil.move(m, 'frame_mean_transformation_mats')

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["average_mats"] = (
            os.getcwd() + '/frame_mean_transformation_mats')

        return outputs


class PetCorrectionFactorInputSpec(BaseInterfaceInputSpec):

    timestamps = Directory(exists=True, desc='Frame start times as detected'
                           'by the motion framing pipeline')


class PetCorrectionFactorOutputSpec(TraitedSpec):

    corr_factors = File(exists=True, desc='Pet correction factors.')


class PetCorrectionFactor(BaseInterface):

    input_spec = PetCorrectionFactorInputSpec
    output_spec = PetCorrectionFactorOutputSpec

    def _run_interface(self, runtime):

        frame_st = np.loadtxt(
            self.inputs.timestamps + '/frame_start_times_4PET.txt', dtype=str)
        start = dt.datetime.strptime(frame_st[0], '%H%M%S.%f')
        end = dt.datetime.strptime(frame_st[-1], '%H%M%S.%f')
        total_duration = (end - start).total_seconds()
        corr_factors = []
        for t in range(len(frame_st) - 1):
            start = dt.datetime.strptime(frame_st[t], '%H%M%S.%f')
            end = dt.datetime.strptime(frame_st[t + 1], '%H%M%S.%f')
            d = (end - start).total_seconds()
            corr_factors.append(d / total_duration)

        with open('correction_factors_PET_data.txt', 'w') as f:
            for el in corr_factors:
                f.write(str(el) + '\n')
            f.close()

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["corr_factors"] = (
            os.getcwd() + '/correction_factors_PET_data.txt')

        return outputs


class UmapAlign2ReferenceInputSpec(BaseInterfaceInputSpec):

    average_mats = Directory(exists=True, desc='directory with all the average'
                             ' transformation matrices for each detected '
                             'frame.')
    ute_regmat = File(exists=True, desc='registration mat between ute image '
                      'and reference.')
    ute_qform_mat = File(exists=True, desc='qform mat between ute and '
                         'reference.')
    umap = File(exists=True, desc='If a umap is provided, the function will '
                'align it to the head position in each frame. Default is None',
                default=None)
    pct = traits.Bool(desc='if True, the function will assume that the '
                      'provided umap is continuos values, as the pseudo CT '
                      'umap. Otherwise, it will assume that the values are '
                      'discrete. Default is False.')


class UmapAlign2ReferenceOutputSpec(TraitedSpec):

    umaps_align2ref = Directory(desc='directory with all the realigned umaps '
                                '(if a umaps is provided as input).')


class UmapAlign2Reference(BaseInterface):

    input_spec = UmapAlign2ReferenceInputSpec
    output_spec = UmapAlign2ReferenceOutputSpec

    def _run_interface(self, runtime):

        average_mats = sorted(glob.glob(self.inputs.average_mats + '/*.txt'))
        umap = self.inputs.umap
        pct = self.inputs.pct
        ute_regmat = self.inputs.ute_regmat
        ute_qform_mat = self.inputs.ute_qform_mat
        outname = 'Frame'

        for i, mat in enumerate(average_mats):
            self.UmapAlign2Reference_calc(mat, i, ute_regmat, ute_qform_mat,
                                          outname, umap, pct=pct)

        umaps = glob.glob('Frame_*_umap.nii.gz')
        if os.path.isdir('umaps_align2ref') is False:
            os.mkdir('umaps_align2ref')
        for u in umaps:
            shutil.move(u, 'umaps_align2ref')

        return runtime

    def UmapAlign2Reference_calc(self, mat, i, ute_regmat, ute_qform_mat,
                                 outname, umap, pct=False):

        mat = np.loadtxt(mat)
        utemat = np.loadtxt(ute_regmat)
        utemat_qform = np.loadtxt(ute_qform_mat)
        utemat_qform_inv = np.linalg.inv(utemat_qform)
        ute2frame = np.dot(mat, utemat)
        ute2frame_qform = np.dot(utemat_qform_inv, ute2frame)
        ute2frame_qform_inv = np.linalg.inv(ute2frame_qform)

        np.savetxt('{0}_{1}_ref_to_ute.mat'.format(outname, str(i).zfill(3)),
                   ute2frame_qform)
        np.savetxt(
            '{0}_{1}_ref_to_ute_inv.mat'.format(outname, str(i).zfill(3)),
            ute2frame_qform_inv)

        if pct:
            interp = 'trilinear'
        else:
            interp = 'nearestneighbour'
        flt = fsl.FLIRT(bins=256, cost_func='corratio')
        flt.inputs.reference = umap
        flt.inputs.in_file = umap
        flt.inputs.interp = interp
        flt.inputs.in_matrix_file = ('Frame_{0}_ref_to_ute.mat'
                                     .format(str(i).zfill(3)))
        flt.inputs.out_file = 'Frame_{0}_umap.nii.gz'.format(
            str(i).zfill(3))
        flt.inputs.apply_xfm = True
        flt.run()

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["umaps_align2ref"] = (
            os.getcwd() + '/umaps_align2ref')

        return outputs


class CreateMocoSeriesInputSpec(BaseInterfaceInputSpec):

    moco_template = File(exists=True, mandatory=True, desc='Existing moco '
                         'series template to modify according to the extracted'
                         'motion information.')
    motion_par = File(exists=True, mandatory=True, desc='Text file with the '
                      'motion parameters extracted by the mean displacement '
                      'calculation pipeline.')
    start_times = File(exists=True, mandatory=True, desc='start times of all '
                       'the sequences (or volumes) acquired in the study ('
                       'this is the output of the mean displacement calculatio'
                       'n pipeline).')


class CreateMocoSeriesOutputSpec(TraitedSpec):

    modified_moco = Directory(entries=True, desc='Directory with the new moco '
                              'series')


class CreateMocoSeries(BaseInterface):

    input_spec = CreateMocoSeriesInputSpec
    output_spec = CreateMocoSeriesOutputSpec

    def _run_interface(self, runtime):

        moco_template = self.inputs.moco_template
        motion_par = np.loadtxt(self.inputs.motion_par)
        start_times = np.loadtxt(self.inputs.start_times, dtype=str)
        start_times = start_times[:-1]

        if len(motion_par) != len(start_times):
            raise Exception('Detected a different number of motion parameters '
                            'and start times. This number must be the same in '
                            'order to create a new moco series. Please check.')
        motion_par_moco = [self.fsl2moco(x) for x in motion_par]
        new_uid = pydicom.uid.generate_uid()
        for i in range(len(start_times)):
            hd = pydicom.read_file(moco_template)
            for n in range(3):
                hd[0x19, 0x1025].value[n] = motion_par_moco[i][n]
            for n in range(3):
                hd[0x19, 0x1026].value[n] = motion_par_moco[i][n + 3]
            hd.AcquisitionTime = start_times[i]
            hd.InstanceNumber = pydicom.valuerep.IS(i + 1)
            hd.AcquisitionNumber = pydicom.valuerep.IS(i + 1)
            hd.SeriesInstanceUID = new_uid
            hd.SeriesDescription = 'MoCoSeries'
            hd.SeriesNumber = '150'
            hd.save_as('{}.IMA'.format(str(i).zfill(6)))

        os.mkdir('new_moco_series')
        dicoms = sorted(glob.glob('*.IMA'))
        for dcm in dicoms:
            shutil.move(dcm, 'new_moco_series')

        return runtime

    def fsl2moco(self, mp):
        rot_x_moco = -self.rad2degree(mp[1])
        rot_y_moco = self.rad2degree(mp[0])
        rot_z_moco = -self.rad2degree(mp[2])
        trans_x_moco = -mp[4]
        trans_y_moco = mp[3]
        trans_z_moco = -mp[5]
        return [trans_x_moco, trans_y_moco, trans_z_moco, rot_x_moco,
                rot_y_moco, rot_z_moco]

    def rad2degree(self, alpha_rad):
        return alpha_rad * 180 / np.pi

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["modified_moco"] = os.getcwd() + '/new_moco_series'

        return outputs


class FixedBinningInputSpec(BaseInterfaceInputSpec):

    n_frames = traits.Int(desc='Number of frames you want to have '
                          'realignment matrices for.')
    pet_offset = traits.Int(desc='seconds from the start of the PET you want '
                            'to discard before starting the data binning.')
    bin_len = traits.Int(desc='Temporal length in seconds for each bin.')
    start_times = File(desc='Start times of all the scans in the study. This '
                       'is generated by mean displacement calculation '
                       'pipeline.')
    pet_duration = traits.Int(desc='PET temporal duration in seconds.')
    pet_start_time = traits.Str(desc='PET start time')
    motion_mats = File(exists=True, desc='Text file with the list of all the '
                       'motion matrices.')


class FixedBinningOutputSpec(TraitedSpec):

    average_bin_mats = Directory(desc='Directory with all the matrices to be '
                                 'used to realign the reconstructed '
                                 'fixed-binning PET images.')


class FixedBinning(BaseInterface):

    input_spec = FixedBinningInputSpec
    output_spec = FixedBinningOutputSpec

    def _run_interface(self, runtime):

        n_frames = self.inputs.n_frames
        pet_offset = self.inputs.pet_offset
        bin_len = self.inputs.bin_len
        start_times = np.loadtxt(self.inputs.start_times, dtype=str)
        pet_duration = self.inputs.pet_duration
        pet_start_time = self.inputs.pet_start_time
        motion_mats = np.loadtxt(self.inputs.motion_mats, dtype=str)
        if n_frames == 0 and pet_offset == 0:
            pet_len = pet_duration
        elif n_frames == 0 and pet_offset != 0:
            pet_len = pet_duration - pet_offset
        elif n_frames != 0:
            pet_len = bin_len * n_frames

        MR_start_time = dt.datetime.strptime(str(start_times[0]), '%H%M%S.%f')
        start_times_diff = [
            (dt.datetime.strptime(str(start_times[i + 1]), '%H%M%S.%f') -
             dt.datetime.strptime(
                 str(start_times[i]), '%H%M%S.%f')).total_seconds()
            for i in range(len(start_times) - 1)]
        scan_duration = np.cumsum(np.asarray(start_times_diff))

        pet_st = (dt.datetime.strptime(pet_start_time, '%H%M%S.%f') +
                  dt.timedelta(seconds=pet_offset))
        PetBins = [pet_st + dt.timedelta(seconds=x) for x in
                   range(0, pet_len, bin_len)]
        MrBins = [MR_start_time + dt.timedelta(seconds=x)
                  for x in scan_duration]
        MrStartPoints = [MrBins[i] + dt.timedelta(
            seconds=(MrBins[i + 1] - MrBins[i]).total_seconds() / 2) for i in
                         range(len(MrBins) - 1)]

        indxs = []
        PetBins.append(pet_st + dt.timedelta(seconds=pet_len))
        if pet_offset != 0:
            print(('PET start time offset of {0} seconds detected. '
                   'Fixed binning will start at {2} and will last '
                   'for {1} seconds.'.format(str(pet_offset), str(pet_len),
                                             pet_st.strftime('%H%M%S.%f'))))
        for pet_bin in PetBins:

            for i in range(len(MrStartPoints) - 1):
                if (pet_bin > MrStartPoints[i] and
                        pet_bin < MrStartPoints[i + 1]):
                    MrDiff = (
                        (MrStartPoints[i + 1] - MrStartPoints[i]).total_seconds())
                    w0 = (MrStartPoints[i + 1] - pet_bin).total_seconds() / MrDiff
                    w1 = (pet_bin - MrStartPoints[i]).total_seconds() / MrDiff
                    indxs.append([[w0, i], [w1, i + 1]])
                    break
                elif pet_bin < MrStartPoints[i]:
                    indxs.append([[1, i], [0, i + 1]])
                    break
        while len(indxs) < len(PetBins):
            indxs.append([[0, len(MrStartPoints) - 2],
                          [1, len(MrStartPoints) - 1]])
        z = 0
        for ii in range(len(indxs) - 1):
            start = indxs[ii]
            end = indxs[ii + 1]
            s1 = start[0][1]
            e1 = start[1][1]
            s2 = end[0][1]
            e2 = end[1][1]
            if s1 == s2 and e1 == e2:
                mat_s1 = np.loadtxt(motion_mats[s1])
                mat_e1 = np.loadtxt(motion_mats[e1])
                av_mat = start[0][0] * mat_s1 + start[1][0] * mat_e1
                np.savetxt(
                    'average_motion_mat_bin_{0}.txt'.format(str(z).zfill(3)),
                    av_mat)
                z = z + 1
            elif (s1 + 1 == s2 and e1 + 1 == e2) or (s1 + 2 == s2 and e1 + 2 == e2):
                mat_s1 = np.loadtxt(motion_mats[s1])
                mat_e1 = np.loadtxt(motion_mats[e1])
                mat_s2 = np.loadtxt(motion_mats[s2])
                mat_e2 = np.loadtxt(motion_mats[e2])
                av_mat_1 = start[0][0] * mat_s1 + start[1][0] * mat_e1
                av_mat_2 = end[0][0] * mat_s2 + end[1][0] * mat_e2
                mean_mat = (av_mat_1 + av_mat_2) / 2
                np.savetxt(
                    'average_motion_mat_bin_{0}.txt'.format(str(z).zfill(3)),
                    mean_mat)
                z = z + 1
            else:
                mat_tot = np.zeros((4, 4, (s2 - s1)))
                mat_s1 = np.loadtxt(motion_mats[s1])
                mat_e1 = np.loadtxt(motion_mats[e1])
                mat_s2 = np.loadtxt(motion_mats[s2])
                mat_e2 = np.loadtxt(motion_mats[e2])
                mat_tot[:, :, 0] = (
                    start[0][0] * mat_s1 + start[1][0] * mat_e1)
                mat_tot[:, :, -1] = (
                    end[0][0] * mat_s2 + end[1][0] * mat_e2)
                for i, m in enumerate(range(e1 + 1, s2)):
                    mat_tot[:, :, i + 1] = np.loadtxt(motion_mats[m])
                mean_mat = np.mean(mat_tot, axis=2)
                np.savetxt(
                    'average_motion_mat_bin_{0}.txt'
                    .format(str(z).zfill(3)), mean_mat)
                z = z + 1
        os.mkdir('average_bin_mats')
        files = glob.glob('*bin*.txt')
        for f in files:
            shutil.move(f, 'average_bin_mats')

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["average_bin_mats"] = os.getcwd() + '/average_bin_mats'

        return outputs


class ReorientUmapInputSpec(BaseInterfaceInputSpec):

    umap = Directory(exists=True)
    niftis = traits.List()


class ReorientUmapOutputSpec(TraitedSpec):

    reoriented_umaps = traits.List()


class ReorientUmap(BaseInterface):
    """Had to write this pipeline because I found out that sometimes the
    reoriented umaps (in nifti format) had different orientation with respect
    to the original umap (dicom). This pipeline uses mrconvert to extract the
    strides from the dicom umap and to apply them to each of the nifti umaps.
    """
    input_spec = ReorientUmapInputSpec
    output_spec = ReorientUmapOutputSpec

    def _run_interface(self, runtime):

        niftis = self.inputs.niftis
        umap = self.inputs.umap

        for nifti in niftis:
            _, base, ext = split_filename(nifti)
            outname = base + '_reorient' + ext
            cmd = ('mrconvert -strides {0} {1} {2}'
                   .format(umap, nifti, outname))
            sp.check_output(cmd, shell=True)

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["reoriented_umaps"] = sorted(glob.glob(
            os.getcwd() + '/*_reorient.*'))

        return outputs



from nipype.interfaces.base import traits, File, TraitedSpec
from nipype.interfaces.mrtrix3.reconst import (
    MRTrix3Base, MRTrix3BaseInputSpec)


class GlobalTractographyInputSpec(MRTrix3BaseInputSpec):
    in_file = File(exists=True, argstr='%s', mandatory=True,
                    position=-3, desc='input dMRI file')
    wm_response = File(exists=True, argstr='%s', mandatory=True,
                       position=-2)
    out_file = File(name_template='%s_tracks.tck',
                    name_source='in_file',
                    argstr='%s',
                    desc=("Output streamlines"), position=-1)
    num_iterations = traits.Int(1e9, argstr='-niter %s',
                                usedefault=True,
                                desc="number of tracks to generate")
    mask = File(exists=True, argstr='-mask %s', mandatory=False,
                desc="Brain mask")


class GlobalTractographyOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc=("Output streamlines"))


class GlobalTractography(MRTrix3Base):
    """Global tractography"""

    _cmd = "tckglobal"
    input_spec = GlobalTractographyInputSpec
    output_spec = GlobalTractographyOutputSpec



import numpy as np
from itertools import chain, repeat
from copy import copy
import os.path as op
import matplotlib.image
import tempfile
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import subprocess as sp
from banana.exceptions import BananaUsageError
from banana.file_format import nifti_gz_format, nifti_format, dicom_format


class ImageDisplayMixin():

    def display_slice_panel(self, filesets, img_size=5,
                            row_kwargs=None, offset=None, **kwargs):
        """
        Displays an image in a Nx3 panel axial, coronal and sagittal
        slices for the filesets correspdong to each of the data names
        provided.

        Parameters
        ----------
        data_names : List[str]
            List of image names to plot as rows of a panel
        size : Tuple(2)[int]
            Size of the figure to plot
        row_kargs : List[Dict[str, *]]
            A list of row-specific kwargs to passed on to
            _display_mid_slices
        offset : Tuple(3)[int]
            An array of integers with which to offset the slices displayed
        """
        n_rows = len(filesets)
        if row_kwargs is None:
            row_kwargs = repeat({})
        elif not n_rows == len(row_kwargs):
            raise BananaUsageError(
                "Length of row_kwargs ({}) needs to "
                "match length of filesets ({})"
                .format(len(row_kwargs), n_rows))
        # Set up figure
        gs = GridSpec(n_rows, 3)
        gs.update(wspace=0.0, hspace=0.0)
        fig = plt.figure(figsize=(3 * img_size,
                                  n_rows * img_size))
        # Loop through derivatives and generate image
        for i, (fileset, rkwargs) in enumerate(zip(filesets,
                                                   row_kwargs)):
            array = fileset.get_array()
            header = fileset.get_header()
            if fileset.format in (nifti_format, nifti_gz_format):
                vox = header['pixdim'][1:4]
            elif fileset.format == dicom_format:
                vox = [float(v) for v in header.PixelSpacing]
                vox.append(float(header.SliceThickness))
            else:
                raise BananaUsageError(
                    "'{}' format images are not supported for display slice "
                    .format(fileset.format))
            rkwargs = copy(rkwargs)
            rkwargs.update(kwargs)
            try:
                self._display_mid_slices(array, vox, fig, gs, i,
                                         offset=offset, **rkwargs)
            except BananaUsageError as e:
                raise BananaUsageError(
                    str(e) + " displaying {}".format(fileset.path))
        # Remove space around figure
        plt.tight_layout(0.0)
        # Either show image or save it to file

    def display_tcks_with_mrview(self, tcks, backgrounds, padding=1,
                                  img_size=5, offset=None):
        """
        Displays dMRI tractography streamlines using MRtrix's mrview to
        display and screenshot to file the streamlines. Then reload,
        crop and combine the different slice orientations into
        a single panel.

        Parameters
        ----------
        save_path : str | None
            The path of the file to save the image at. If None the
            images are displayed instead
        size : Tuple(2)[int]
            The size of the combined figure to plot
        offset : Tuple(3)[int]
            An array of integers with with to offset the slices displayed
        """
        n_rows = len(tcks)
        # Create figure in which to aggregate the plots
        gs = GridSpec(n_rows, 3)
        gs.update(wspace=0.0, hspace=0.0)
        fig = plt.figure(figsize=(3 * img_size,
                                  n_rows * img_size))
        # Create temp dir to store screen captures
        tmpdir = tempfile.mkdtemp()
        for i, (tck, bg) in enumerate(zip(tcks, backgrounds)):
            # Call mrview to display the tracks, capture them to
            # file and then reload the image into a matplotlib
            # grid
            options = ['-tractography.load', tck.path, '-noannotations']
            if offset is not None:
                array = bg.get_array()
                centre = self.image_centre(array, offset)
                options.extend(['-voxel', '{},{},{}'.format(*centre)])
            # Set options to remove cursor, capture hte image and exit
            options.extend(['-lock', 'yes', '-capture.grab', '-exit'])
            imgs = []
            for i in range(3):
                sp.call(
                    '{} {} -plane {} {}'.format(
                        self.mrview_path, bg.path, 2 - i,
                        ' '.join(options)), cwd=tmpdir,
                    shell=(self.mrview_path == 'mrview'))
                img = matplotlib.image.imread(
                    op.join(tmpdir, 'screenshot0000.png'))
                imgs.append(self.crop(img, border=padding))
            padded_size = max(chain(*(a.shape for a in imgs)))
            for i, img in enumerate(imgs):
                img = self.pad_to_size(img, (padded_size,
                                              padded_size))
                axis = fig.add_subplot(gs[i])
                axis.get_xaxis().set_visible(False)
                axis.get_yaxis().set_visible(False)
                plt.imshow(img)
        plt.tight_layout(0.0)

    def show(self, path=None, subj_id=None, visit_id=None):
        """
        Save current figure to file or show it
        """
        if path is None:
            plt.show()
        else:
            base, ext = op.splitext(path)
            if len(list(self.subject_ids)) > 1:
                base += '-sub{}'.format(subj_id)
            if len(list(self.visit_ids)) > 1:
                base += '-vis{}'.format(visit_id)
            plt.savefig(base + ext)

    def _display_mid_slices(self, array, vox_sizes, fig, grid_spec,
                            row_index, padding=1, vmax=None, vmin=None,
                            vmax_percentile=98, vmin_percentile=2,
                            offset=None):
        # Guard agains NaN
        array[np.isnan(array)] = 0.0
        # Crop black-space around array
        array = self.crop(array, padding)
        # Pad out image array into cube
        padded_size = np.max(array.shape)
        # Get centre of the image in order to take image slices there
        centre = self.image_centre(array, offset)
        # Get dynamic range of array
        if vmax is None:
            assert vmax_percentile is not None
            vmax = np.percentile(array, vmax_percentile)
        if vmin is None:
            assert vmin_percentile is not None
            vmin = np.percentile(array, vmin_percentile)

        # Function to plot a slice
        def display_slice(slce, index, aspect):
            axis = fig.add_subplot(grid_spec[index])
            axis.get_xaxis().set_visible(False)
            axis.get_yaxis().set_visible(False)
            pad_vert = int(np.round(padded_size / aspect))
            padded_slce = self.pad_to_size(slce, (pad_vert, padded_size))
            plt.imshow(padded_slce,
                       interpolation='bilinear',
                       cmap='gray', aspect=aspect,
                       vmin=vmin, vmax=vmax)
        # Display slices
        display_slice(np.squeeze(array[-1:0:-1, -1:0:-1, centre[2]]).T,
                      row_index * 3, vox_sizes[1] / vox_sizes[0])
        display_slice(np.squeeze(array[-1:0:-1, centre[1], -1:0:-1]).T,
                      row_index * 3 + 1, vox_sizes[2] / vox_sizes[0])
        display_slice(np.squeeze(array[centre[0], -1:0:-1, -1:0:-1]).T,
                      row_index * 3 + 2, vox_sizes[2] / vox_sizes[1])

    @classmethod
    def crop(cls, array, border=0):
        """
        Crops an image array to the nonzero voxels (+ equal border in
        each dimension)
        """
        nonzeros = np.argwhere(array)
        # Build index slices for each dimension of the array
        slices = []
        for i, (a, b) in enumerate(zip(nonzeros.min(axis=0),
                                       nonzeros.max(axis=0))):
            # Append the index slice for the
            slices.append(slice(max(a - border, 0),
                                min(b + border, array.shape[i])))
        return array[tuple(slices)]

    @classmethod
    def image_centre(cls, array, offset=None):
        """
        Returns the centre point of the non-zero voxels in the image
        """
        nonzeros = np.argwhere(array)
        min_ind = nonzeros.min(axis=0)
        max_ind = nonzeros.max(axis=0)
        centre = ((max_ind - min_ind) // 2) + min_ind
        if offset is not None:
            centre += np.array(offset, dtype=int)
            if np.any(centre < 0) or np.any(centre > array.shape):
                raise BananaUsageError(
                    "Specified offset ({}) is larger than the "
                    "dimension of the image / 2 ({})"
                    .format(offset, array.shape // 2))
        return centre

    @classmethod
    def pad_to_size(cls, array, size):
        pad_before = [
            (size[i] - array.shape[i]) // 2
            for i in range(2)]
        pad_after = [
            size[i] - array.shape[i] - pad_before[i]
            for i in range(2)]
        padding = list(zip(pad_before, pad_after))
        padding.extend((0, 0) for _ in range(array.ndim - 2))
        return np.pad(array, padding, 'constant')

    @property
    def mrview_path(self):
        """
        The path to the 'mrview' executable. Returns simply 'mrview'
        if it hasn't been explicitly, which assumes it is on the path
        """
        try:
            return self._mrview_path
        except AttributeError:
            return 'mrview'

    @mrview_path.setter
    def mrview_path(self, path):
        self._mrview_path = path



from copy import copy
from nipype.interfaces.freesurfer.preprocess import ReconAll
# from arcana.utils.interfaces import DummyReconAll as ReconAll
from banana.requirement import freesurfer_req, ants_req, fsl_req
from banana.citation import freesurfer_cites, fsl_cite
from nipype.interfaces import fsl, ants
from arcana.utils.interfaces import Merge
from banana.file_format import (
    nifti_gz_format, nifti_format, zip_format, STD_IMAGE_FORMATS,
    directory_format)
from arcana.data import FilesetSpec, FilesetInputSpec
from arcana.utils.interfaces import JoinPath
from .t2 import T2Study, MriStudy
from arcana.study.base import StudyMetaClass
from arcana.study import ParamSpec, SwitchSpec
from banana.reference import LocalReferenceData


class T1Study(T2Study, metaclass=StudyMetaClass):

    add_data_specs = [
        FilesetSpec('fs_recon_all', zip_format, 'freesurfer_pipeline'),
        FilesetSpec('brain', nifti_gz_format, 'brain_extraction_pipeline'),
        FilesetInputSpec(
            't2_coreg', STD_IMAGE_FORMATS, optional=True,
            desc=("A coregistered T2 image to use in freesurfer to help "
                  "distinguish the peel surface")),
        # Templates
        FilesetInputSpec('suit_mask', STD_IMAGE_FORMATS,
                         frequency='per_study',
                         default=LocalReferenceData('SUIT', nifti_format))]

    add_param_specs = [
        SwitchSpec('bet_method', 'fsl_bet',
                   choices=MriStudy.parameter_spec('bet_method').choices),
        SwitchSpec('bet_robust', False),
        SwitchSpec('bet_reduce_bias', True),
        ParamSpec('bet_f_threshold', 0.1),
        ParamSpec('bet_g_threshold', 0.0)]
#         SwitchSpec('bet_method', 'optibet',
#                    choices=MriStudy.parameter_spec('bet_method').choices),
#         SwitchSpec('bet_robust', True),
#         ParamSpec('bet_f_threshold', 0.57),
#         ParamSpec('bet_g_threshold', -0.1)]

    def freesurfer_pipeline(self, **name_maps):
        """
        Segments grey matter, white matter and CSF from T1 images using
        SPM "NewSegment" function.

        NB: Default values come from the W2MHS toolbox
        """
        pipeline = self.new_pipeline(
            name='segmentation',
            name_maps=name_maps,
            desc="Segment white/grey matter and csf",
            citations=copy(freesurfer_cites))

        # FS ReconAll node
        recon_all = pipeline.add(
            'recon_all',
            interface=ReconAll(
                directive='all',
                openmp=self.processor.num_processes),
            inputs={
                'T1_files': ('preproc', nifti_gz_format)},
            requirements=[freesurfer_req.v('5.3')],
            wall_time=2000)

        if self.provided('t2_coreg'):
            pipeline.connect_input('t2_coreg', recon_all, 'T2_file',
                                   nifti_gz_format)
            recon_all.inputs.use_T2 = True

        # Wrapper around os.path.join
        pipeline.add(
            'join',
            JoinPath(),
            inputs={
                'dirname': (recon_all, 'subjects_dir'),
                'filename': (recon_all, 'subject_id')},
            outputs={
                'fs_recon_all': ('path', directory_format)})

        return pipeline

    def segmentation_pipeline(self, **name_maps):
        pipeline = super(T1Study, self).segmentation_pipeline(img_type=1,
                                                              **name_maps)
        return pipeline

    def bet_T1(self, **name_maps):

        pipeline = self.new_pipeline(
            name='BET_T1',
            name_maps=name_maps,
            desc=("python implementation of BET"),
            citations=[fsl_cite])

        bias = pipeline.add(
            'n4_bias_correction',
            ants.N4BiasFieldCorrection(),
            inputs={
                'input_image': ('t1', nifti_gz_format)},
            requirements=[ants_req.v('1.9')],
            wall_time=60, mem_gb=12)

        pipeline.add(
            'bet',
            fsl.BET(frac=0.15, reduce_bias=True),
            inputs={
                'in_file': (bias, 'output_image')},
            outputs={
                'betted_T1': ('out_file', nifti_gz_format),
                'betted_T1_mask': ('mask_file', nifti_gz_format)},
            requirements=[fsl_req.v('5.0.8')], mem_gb=8,
            wall_time=45)

        return pipeline

    def cet_T1(self, **name_maps):
        pipeline = self.new_pipeline(
            name='CET_T1',
            name_maps=name_maps,
            desc=("Construct cerebellum mask using SUIT template"),
            citations=[fsl_cite])

        # FIXME: Should convert to inputs
        nl = self._lookup_nl_tfm_inv_name('MNI')
        linear = self._lookup_l_tfm_to_name('MNI')

        # Initially use MNI space to warp SUIT into T1 and threshold to mask
        merge_trans = pipeline.add(
            'merge_transforms',
            Merge(2),
            inputs={
                'in2': (nl, nifti_gz_format),
                'in1': (linear, nifti_gz_format)})

        apply_trans = pipeline.add(
            'ApplyTransform',
            ants.resampling.ApplyTransforms(
                interpolation='NearestNeighbor',
                input_image_type=3,
                invert_transform_flags=[True, False]),
            inputs={
                'reference_image': ('betted_T1', nifti_gz_format),
                'input_image': ('suit_mask', nifti_gz_format),
                'transforms': (merge_trans, 'out')},
            requirements=[ants_req.v('1.9')], mem_gb=16,
            wall_time=120)

        pipeline.add(
            'maths2',
            fsl.utils.ImageMaths(
                suffix='_optiBET_cerebellum',
                op_string='-mas'),
            inputs={
                'in_file': ('betted_T1', nifti_gz_format),
                'in_file2': (apply_trans, 'output_image')},
            outputs={
                'cetted_T1': ('out_file', nifti_gz_format),
                'cetted_T1_mask': ('output_image', nifti_gz_format)},
            requirements=[fsl_req.v('5.0.8')], mem_gb=16,
            wall_time=5)

        return pipeline



from copy import deepcopy, copy
from abc import ABCMeta, abstractmethod
import os
import os.path as op
import json
import pydicom
import numpy as np
from arcana.data.file_format import FileFormat, Converter
from banana.interfaces.mrtrix import MRConvert
from banana.requirement import (
    dcm2niix_req, mrtrix_req)
from banana.interfaces.converters import Dcm2niix  # @UnusedImport
from banana.exceptions import BananaUsageError
import nibabel
# Import base file formats from Arcana for convenience
from arcana.data.file_format import (
    text_format, directory_format, zip_format, targz_format,  # @UnusedImport
    png_format, jpg_format, gif_format,  # @UnusedImport
    UnzipConverter, UnTarGzConverter, IdentityConverter)  # @UnusedImport


class Dcm2niixConverter(Converter):

    interface = Dcm2niix(compression='y')
    input = 'input_dir'
    output = 'converted'
    requirements = [dcm2niix_req.v('1.0.2')]


class MrtrixConverter(Converter):

    input = 'in_file'
    output = 'out_file'
    requirements = [mrtrix_req.v(3)]

    @property
    def interface(self):
        return MRConvert(
            out_ext=self.output_format.extension,
            quiet=True)


# =====================================================================
# Custom loader functions for different image types
# =====================================================================


class ImageFormat(FileFormat, metaclass=ABCMeta):

    @abstractmethod
    def get_header(self, fileset):
        """
        Returns array data associated with the given path for the
        file format
        """

    @abstractmethod
    def get_array(self, fileset):
        """
        Returns header data associated with the given path for the
        file format
        """

    def contents_equal(self, fileset, other_fileset, rms_tol=None, **kwargs):
        """
        Test whether the (relevant) contents of two image filesets are equal
        given specific criteria

        Parameters
        ----------
        fileset : Fileset
            One of the two filesets to compare
        other_fileset : Fileset
            The other fileset to compare
        rms_tol : float
            The root-mean-square tolerance that is acceptable between the array
            data for the images to be considered equal
        """
        if other_fileset.format != self:
            return False
        if self.headers_diff(fileset, other_fileset, **kwargs):
            return False
        if rms_tol:
            rms_diff = self.rms_diff(fileset, other_fileset)
            return (rms_diff < rms_tol)
        else:
            return np.array_equiv(fileset.get_array(),
                                  other_fileset.get_array())

    def headers_diff(self, fileset, other_fileset, include_keys=None,
                     ignore_keys=None, **kwargs):  # @UnusedVariable
        """
        Check headers to see if all values
        """
        diff = []
        hdr = fileset.get_header()
        hdr_keys = set(hdr.keys())
        other_hdr = other_fileset.get_header()
        if include_keys is not None:
            if ignore_keys is not None:
                raise BananaUsageError(
                    "Doesn't make sense to provide both 'include_keys' ({}) "
                    "and ignore_keys ({}) to headers_equal method"
                    .format(include_keys, ignore_keys))
            include_keys &= hdr_keys
        elif ignore_keys is not None:
            include_keys = hdr_keys - set(ignore_keys)
        else:
            if hasattr(self, 'INCLUDE_HDR_KEYS'):
                if hasattr(self, 'IGNORE_HDR_KEYS'):
                    raise BananaUsageError(
                        "Doesn't make sense to have both 'INCLUDE_HDR_FIELDS'"
                        "and 'IGNORE_HDR_FIELDS' class attributes of class {}"
                        .format(type(self).__name__))
                include_keys = self.INCLUDE_HDR_KEYS
            elif hasattr(self, 'IGNORE_HDR_KEYS'):
                include_keys = hdr_keys - set(self.IGNORE_HDR_KEYS)
            else:
                include_keys = hdr_keys
        for key in include_keys:
            value = hdr[key]
            try:
                other_value = other_hdr[key]
            except KeyError:
                diff.append(key)
            else:
                if isinstance(value, np.ndarray):
                    if not isinstance(other_value, np.ndarray):
                        diff.append(key)
                    else:
                        try:
                            if not np.allclose(value, other_value,
                                               equal_nan=True):
                                diff.append(key)
                        except TypeError:
                            # Fallback to a straight comparison for some dtypes
                            if value != other_value:
                                diff.append(key)
                elif value != other_value:
                    diff.append(key)
        return diff

    def rms_diff(self, fileset, other_fileset):
        """
        Return the RMS difference between the image arrays
        """
        return np.sqrt(np.sum((fileset.get_array() -
                               other_fileset.get_array()) ** 2))


class NiftiFormat(ImageFormat):

    def get_header(self, fileset):
        return dict(nibabel.load(fileset.path).header)

    def get_array(self, fileset):
        return nibabel.load(fileset.path).get_data()

    def get_vox_sizes(self, fileset):
        # FIXME: This won't work for 4-D files
        return self.get_header(fileset)['pixdim'][1:4]

    def get_dims(self, fileset):
        # FIXME: This won't work for 4-D files
        return self.get_header(fileset)['dim'][1:4]


class NiftixFormat(NiftiFormat):

    def get_header(self, fileset):
        hdr = super().get_header(fileset)
        with open(fileset.aux_file('json')) as f:
            hdr.update(json.load(f))
        return hdr


class DicomFormat(ImageFormat):

    SERIES_NUMBER_TAG = ('0020', '0011')

    def dcm_files(self, fileset):
        return [f for f in os.listdir(fileset.path) if f.endswith('.dcm')]

    def get_array(self, fileset):
        image_stack = []
        for fname in self.dcm_files(fileset):
            image_stack.append(
                pydicom.dcmread(op.join(fileset.path, fname)).pixel_array)
        return np.asarray(image_stack)

    def get_header(self, fileset, index=0):
        dcm_files = [f for f in os.listdir(fileset.path) if f.endswith('.dcm')]
        # TODO: Probably should collate fields that vary across the set of
        #       files in the set into lists
        return pydicom.dcmread(op.join(fileset.path, dcm_files[index]))

    def get_vox_sizes(self, fileset):
        hdr = self.get_header(fileset)
        return np.array(hdr.PixelSpacing + [hdr.SliceThickness])

    def get_dims(self, fileset):
        hdr = self.get_header(fileset)
        return np.array((hdr.Rows, hdr.Columns, len(self.dcm_files(fileset))),
                        dtype=int)

    def extract_id(self, fileset):
        return int(fileset.dicom_values([self.SERIES_NUMBER_TAG]))

    def dicom_values(self, fileset, tags):
        """
        Returns a dictionary with the DICOM header fields corresponding
        to the given tag names

        Parameters
        ----------
        tags : List[Tuple[str, str]]
            List of DICOM tag values as 2-tuple of strings, e.g.
            [('0080', '0020')]
        repository_login : <repository-login-object>
            A login object for the repository to avoid having to relogin
            for every dicom_header call.

        Returns
        -------
        dct : Dict[Tuple[str, str], str|int|float]
        """
        try:
            if (fileset._path is None and fileset._repository is not None and
                    hasattr(fileset.repository, 'dicom_header')):
                hdr = fileset.repository.dicom_header(self)
                dct = [hdr[t] for t in tags]
            else:
                # Get the DICOM object for the first file in the fileset
                dcm = fileset.get_header(0)
                dct = [dcm[t].value for t in tags]
        except KeyError as e:
            e.msg = ("{} does not have dicom tag {}".format(
                     self, e.msg))
            raise e
        return dct


class MrtrixImageFormat(ImageFormat):

    def _load_header_and_array(self, fileset):
        with open(fileset.path, 'rb') as f:
            contents = f.read()
        hdr_end = contents.find(b'\nEND\n')
        hdr_contents = contents[:hdr_end].decode('utf-8')
        hdr = dict(l.split(': ', maxsplit=1) for l in hdr_contents.split('\n'))
        for key, value in list(hdr.items()):
            if ',' in value:
                try:
                    hdr[key] = np.array(value.split(','), dtype=int)
                except ValueError:
                    try:
                        hdr[key] = np.array(value.split(','), dtype=float)
                    except ValueError:
                        pass
            else:
                try:
                    hdr[key] = int(value)
                except ValueError:
                    try:
                        hdr[key] = float(value)
                    except ValueError:
                        pass
        del hdr['mrtrix image']  # Delete "magic line" at start of header
        array_start = int(hdr['file'].split()[1])
        array = np.asarray(contents[array_start:])
        dim = [hdr['dim'][int(l[1])] for l in hdr['layout']]
        array = array.reshape(dim)
        return hdr, array

    def get_header(self, fileset):
        self._load_header_and_array(fileset)[0]

    def get_array(self, fileset):
        self._load_header_and_array(fileset)[1]

    def get_vox_sizes(self, fileset):
        return self.get_header(fileset)['vox']

    def get_dims(self, fileset):
        return self.get_header(fileset)['dim']


# =====================================================================
# All Data Formats
# =====================================================================


# NeuroImaging data formats
dicom_format = DicomFormat(name='dicom', extension=None,
                           directory=True, within_dir_exts=['.dcm'],
                           resource_names={'xnat': ['DICOM', 'secondary']})
nifti_gz_x_format = NiftixFormat(name='extended_nifti_gz', extension='.nii.gz',
                                 aux_files={'json': '.json'},
                                 resource_names={'xnat': ['NiFTI_GZ_X',
                                                          'NIFTIX_GZ']})
nifti_format = NiftiFormat(name='nifti', extension='.nii',
                           resource_names={'xnat': ['NiFTI', 'NIFTI']})

nifti_gz_format = NiftiFormat(name='nifti_gz', extension='.nii.gz',
                              resource_names={'xnat': ['NiFTI_GZ',
                                                       'NIFTI_GZ']})
analyze_format = NiftiFormat(name='analyze', extension='.img',
                             aux_files={'header': '.hdr'})
mrtrix_image_format = MrtrixImageFormat(name='mrtrix_image', extension='.mif',
                                        resource_names={'xnat': ['MIF',
                                                                 'MRTRIX']})

# Set converters between image formats

nifti_gz_x_format.set_converter(dicom_format, Dcm2niixConverter)

nifti_format.set_converter(dicom_format, Dcm2niixConverter)
nifti_format.set_converter(analyze_format, MrtrixConverter)
nifti_format.set_converter(nifti_gz_format, MrtrixConverter)
nifti_format.set_converter(mrtrix_image_format, MrtrixConverter)

nifti_gz_format.set_converter(dicom_format, Dcm2niixConverter)
nifti_gz_format.set_converter(nifti_format, MrtrixConverter)
nifti_gz_format.set_converter(analyze_format, MrtrixConverter)
nifti_gz_format.set_converter(mrtrix_image_format, MrtrixConverter)
nifti_gz_format.set_converter(nifti_gz_x_format, IdentityConverter)

analyze_format.set_converter(dicom_format, MrtrixConverter)
analyze_format.set_converter(nifti_format, MrtrixConverter)
analyze_format.set_converter(nifti_gz_format, MrtrixConverter)
analyze_format.set_converter(mrtrix_image_format, MrtrixConverter)

mrtrix_image_format.set_converter(dicom_format, MrtrixConverter)
mrtrix_image_format.set_converter(nifti_format, MrtrixConverter)
mrtrix_image_format.set_converter(nifti_gz_format, MrtrixConverter)
mrtrix_image_format.set_converter(analyze_format, MrtrixConverter)

STD_IMAGE_FORMATS = [dicom_format, nifti_format, nifti_gz_format,
                     nifti_gz_x_format, analyze_format, mrtrix_image_format]

multi_nifti_gz_format = FileFormat(name='multi_nifti_gz', extension=None,
                                   directory=True, within_dir_exts=['.nii.gz'])
multi_nifti_gz_format.set_converter(zip_format, UnzipConverter)
multi_nifti_gz_format.set_converter(targz_format, UnTarGzConverter)

# Tractography formats
mrtrix_track_format = FileFormat(name='mrtrix_track', extension='.tck')

# Tabular formats
rfile_format = FileFormat(name='rdata', extension='.RData')
tsv_format = FileFormat(name='tab_separated', extension='.tsv')
# matlab_format = FileFormat(name='matlab', extension='.mat')
csv_format = FileFormat(name='comma_separated', extension='.csv')
text_matrix_format = FileFormat(name='text_matrix', extension='.mat')

# Diffusion gradient-table data formats
fsl_bvecs_format = FileFormat(name='fsl_bvecs', extension='.bvec')
fsl_bvals_format = FileFormat(name='fsl_bvals', extension='.bval')
mrtrix_grad_format = FileFormat(name='mrtrix_grad', extension='.b')

# Tool-specific formats
eddy_par_format = FileFormat(name='eddy_par',
                             extension='.eddy_parameters')
ica_format = FileFormat(name='ica', extension='.ica', directory=True)
par_format = FileFormat(name='parameters', extension='.par')
motion_mats_format = FileFormat(
    name='motion_mats', directory=True, within_dir_exts=['.mat'],
    desc=("Format used for storing motion matrices produced during "
          "motion detection pipeline"))


# PET formats
list_mode_format = FileFormat(name='pet_list_mode', extension='.bf')

# Raw formats
dat_format = FileFormat(name='dat', extension='.dat')

# MRS format
rda_format = FileFormat(name='raw', extension='.rda')

# Record list of all data formats registered by module (not really
# used currently but could be useful in future)
std_formats = []

# Add all data formats in module to a list of "standard" biomedical formats
for file_format in copy(globals()).values():
    if isinstance(file_format, FileFormat):
        std_formats.append(file_format.name)


# Since the conversion from DICOM->NIfTI is unfortunately slightly
# different between MRConvert and Dcm2niix, these data formats can
# be used in pipeline input specs that need to use MRConvert instead
# of Dcm2niix (i.e. motion-detection pipeline)
mrconvert_nifti_format = deepcopy(nifti_format)
mrconvert_nifti_format.set_converter(dicom_format, MrtrixConverter)
mrconvert_nifti_gz_format = deepcopy(nifti_gz_format)
mrconvert_nifti_gz_format.set_converter(dicom_format, MrtrixConverter)



"""
banana

Copyright (c) 2012-2018 Thomas G. Close, Monash Biomedical Imaging,
Monash University, Melbourne, Australia

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .__about__ import __version__, __authors__
import os

# Should be set explicitly in all FSL interfaces, but this squashes the warning
os.environ['FSLOUTPUTTYPE'] = 'NIFTI_GZ'

# Import all objects from Arcana used to design and apply Banana studies
from arcana import (  # @IgnorePep8
    SubStudySpec, Parameter, ParamSpec, SwitchSpec, FileFormat, Fileset,
    FilesetSpec, FilesetInput, FilesetInputSpec, FilesetCollection, Field,
    FieldSpec, FieldInput, FieldInputSpec, FieldCollection, SingleProc,
    MultiProc, SlurmProc, DEFAULT_PROV_CHECK, DEFAULT_PROV_IGNORE, StaticEnv,
    ModulesEnv, BasicRepo, XnatRepo)

from .bids import BidsRepo  # @IgnorePep8


# Import all Study classes into package root
# from .study.base import (  # @IgnorePep8
#     Study, StudyMetaClass, MultiStudy, MultiStudyMetaClass)



from .base import MriStudy
from arcana.study.base import StudyMetaClass
from arcana.data import FilesetSpec
from nipype.interfaces.fsl.preprocess import FLIRT, ApplyXFM
from nipype.interfaces.fsl.utils import ConvertXFM, Smooth
from nipype.interfaces.fsl.maths import (
    UnaryMaths, BinaryMaths, MultiImageMaths, Threshold)
from nipype.interfaces.spm.preprocess import NewSegment
from nipype.interfaces.utility.base import Select
from banana.interfaces.umap_calc import CoreUmapCalc
from banana.interfaces.mrtrix.utils import MRConvert
from banana.citation import (
    fsl_cite, spm_cite, matlab_cite)
from banana.file_format import (
    dicom_format, nifti_gz_format, nifti_format, text_matrix_format)
from banana.requirement import (
    fsl_req, spm_req, matlab_req)
from arcana.study import SwitchSpec


class UteStudy(MriStudy, metaclass=StudyMetaClass):

    add_data_specs = [
        FilesetSpec('umap', dicom_format),
        FilesetSpec('umap_nifti', nifti_gz_format,
                    'umap_dcm2nii_conversion_pipeline'),
        FilesetSpec('brain', nifti_gz_format, 'brain_extraction_pipeline'),
        FilesetSpec('ute_echo1', dicom_format),
        FilesetSpec('ute_echo2', dicom_format),
        FilesetSpec('umap_ute', dicom_format),
        FilesetSpec('ute1_registered', nifti_gz_format,
                    'registration_pipeline'),
        FilesetSpec('ute2_registered', nifti_gz_format,
                    'registration_pipeline'),
        FilesetSpec('template_to_ute_mat', text_matrix_format,
                    'registration_pipeline'),
        FilesetSpec('ute_to_template_mat', text_matrix_format,
                    'registration_pipeline'),
        FilesetSpec('air_mask', nifti_gz_format,
                    'segmentation_pipeline'),
        FilesetSpec('bones_mask', nifti_gz_format,
                    'segmentation_pipeline'),
        FilesetSpec('sute_cont_template', nifti_gz_format,
                    'umaps_calculation_pipeline'),
        FilesetSpec('sute_fix_template', nifti_gz_format,
                    'umaps_calculation_pipeline'),
        FilesetSpec('sute_fix_ute', nifti_gz_format,
                    'backwrap_to_ute_pipeline'),
        FilesetSpec('sute_cont_ute', nifti_gz_format,
                    'backwrap_to_ute_pipeline')]

    add_param_specs = [
        SwitchSpec('bet_method', 'optibet',
                   choices=MriStudy.parameter_spec('bet_method').choices)]

    template_path = '/home/jakubb/template/template_template0.nii.gz'
    tpm_path = '/environment/packages/spm/12/tpm/head_tpm.nii'

    def header_extraction_pipeline(self, **kwargs):
        return (super(UteStudy, self).
                header_extraction_pipeline_factory(
                    'magnitude', **kwargs))

    def umap_dcm2nii_conversion_pipeline(self, **kwargs):
        return super(UteStudy, self).dcm2nii_conversion_pipeline_factory(
            'umap_dcm2nii', 'umap', **kwargs)

    def registration_pipeline(self, **kwargs):  # @UnusedVariable @IgnorePep8
        """
        Register T1 and T2 to the

        Parameters
        ----------
        """

        pipeline = self.new_pipeline(
            name='registration_pipeline',
            desc="Register ute images to the template",
            citations=(fsl_cite),
            **kwargs)

        echo1_conv = pipeline.add(
            'echo1_conv',
            MRConvert(
                out_ext='.nii.gz'),
            inputs={
                'in_file': ('ute_echo1', dicom_format)},
            outputs={})

        echo2_conv = pipeline.add(
            'echo2_conv',
            MRConvert(
                out_ext='.nii.gz'),
            inputs={
                'in_file': ('ute_echo2', dicom_format)})

        # Create registration node
        registration = pipeline.add(
            'ute1_registration',
            FLIRT(
                reference=self.template_path,
                output_type='NIFTI_GZ',
                searchr_x=[-180, 180],
                searchr_y=[-180, 180],
                searchr_z=[-180, 180],
                bins=256,
                cost_func='corratio'),
            inputs={
                'in_file': (echo1_conv, 'out_file')},
            outputs={
                'ute1_registered': ('out_file', nifti_format),
                'ute_to_template_mat': ('out_matrix_file',
                                        text_matrix_format)},
            requirements=[fsl_req.v('5.0.10')],
            wall_time=180)

        # Inverse matrix conversion
        pipeline.add(
            'inverse_matrix_conversion',
            ConvertXFM(
                invert_xfm=True),
            inputs={
                'in_file': (registration, 'out_matrix_file')},
            outputs={
                'template_to_ute_mat': ('out_file', text_matrix_format)},
            requirements=[fsl_req.v('5.0.10')],
            wall_time=10)

        # UTE_echo_2 transformation
        pipeline.add(
            'transform_t2',
            ApplyXFM(
                output_type='NIFTI_GZ',
                reference=self.template_path,
                apply_xfm=True),
            inputs={
                'in_matrix_file': (registration, 'out_matrix_file'),
                'in_file': (echo2_conv, 'out_file')},
            outputs={
                'ute2_registered': ('out_file', nifti_format)},
            requirements=[fsl_req.v('5.0.10')],
            wall_time=10)

        # Connect outputs

        return pipeline

    def segmentation_pipeline(self, **kwargs):  # @UnusedVariable @IgnorePep8

        pipeline = self.new_pipeline(
            name='ute1_segmentation',
            desc="Segmentation of the first echo UTE image",
            citations=(spm_cite, matlab_cite),
            **kwargs)

        segmentation = pipeline.add(
            'ute1_registered_segmentation',
            NewSegment(
                affine_regularization='none',
                tissues=[
                    ((self.tpm_path, 1), 1, (True, False), (False, False)),
                    ((self.tpm_path, 2), 1, (True, False), (False, False)),
                    ((self.tpm_path, 3), 2, (True, False), (False, False)),
                    ((self.tpm_path, 4), 3, (True, False), (False, False)),
                    ((self.tpm_path, 5), 4, (True, False), (False, False)),
                    ((self.tpm_path, 6), 3, (True, False), (False, False))]),
            inputs={
                'channel_files': ('ute1_registered', nifti_format)},
            requirements=[matlab_req.v('R2015'), spm_req.v('12')],
            wall_time=480)

        select_bones_pm = pipeline.add(
            'select_bones_pm_from_SPM_new_segmentation',
            Select(
                index=3),
            inputs={
                'inlist': (segmentation, 'native_class_images')},
            requirements=[],
            wall_time=5)

        select_air_pm = pipeline.add(
            'select_air_pm_from_SPM_new_segmentation',
            Select(
                index=5),
            inputs={
                'inlist': (segmentation, 'native_class_images')},
            requirements=[],
            wall_time=5)

        threshold_bones = pipeline.add(
            'bones_probabilistic_map_thresholding',
            Threshold(
                output_type="NIFTI_GZ",
                direction='below',
                thresh=0.2),
            inputs={
                'in_file': (select_bones_pm, 'out')},
            requirements=[fsl_req.v('5.0.10')],
            wall_time=5)

        pipeline.add(
            'bones_probabilistic_map_binarization',
            UnaryMaths(
                output_type="NIFTI_GZ",
                operation='bin'),
            inputs={
                'in_file': (threshold_bones, 'out_file')},
            outputs={
                'bones_mask': ('out_file', nifti_format)},
            requirements=[fsl_req.v('5.0.10')],
            wall_time=5)

        threshold_air = pipeline.add(
            'air_probabilistic_maps_thresholding',
            Threshold(
                output_type="NIFTI_GZ",
                direction='below',
                thresh=0.1),
            inputs={
                'in_file': (select_air_pm, 'out')},
            requirements=[fsl_req.v('5.0.10')],
            wall_time=5)

        pipeline.add(
            'air_probabilistic_map_binarization',
            UnaryMaths(
                output_type="NIFTI_GZ",
                operation='bin'),
            inputs={
                'in_file': (threshold_air, 'out_file')},
            outputs={
                'air_mask': ('out_file', nifti_format)},
            requirements=[fsl_req.v('5.0.10')],
            wall_time=5)

        return pipeline

    def umaps_calculation_pipeline(self, **kwargs):

        pipeline = self.new_pipeline(
            name='core_umaps_calculation',
            desc="Umaps calculation in the template space",
            citations=(matlab_cite),
            **kwargs)

        pipeline.add(
            'umaps_calculation_based_on_masks_and_r2star',
            CoreUmapCalc(),
            inputs={
                'ute1_reg': ('ute1_registered', nifti_gz_format),
                'ute2_reg': ('ute2_registered', nifti_gz_format),
                'air__mask': ('air_mask', nifti_gz_format),
                'bones__mask': ('bones_mask', nifti_gz_format)},
            outputs={
                'sute_cont_template': ('sute_cont_template', nifti_gz_format),
                'sute_fix_template': ('sute_fix_template', nifti_gz_format)},
            requirements=[matlab_req.v('R2015')],
            wall_time=20)

        return pipeline

    def backwrap_to_ute_pipeline(self, **kwargs):

        pipeline = self.new_pipeline(
            name='backwrap_to_ute',
            desc="Moving umaps back to the UTE space",
            citations=(matlab_cite),
            **kwargs)

        echo1_conv = pipeline.add(
            'echo1_conv',
            MRConvert(
                out_ext='.nii.gz'),
            inputs={
                'in_file': ('ute_echo1', dicom_format)})

        umap_conv = pipeline.add(
            'umap_conv',
            MRConvert(
                out_ext='.nii.gz'),
            inputs={
                'in_file': ('umap_ute', dicom_format)})

        zero_template_mask = pipeline.add(
            'zero_template_mask',
            BinaryMaths(
                operation="mul",
                operand_value=0,
                output_type='NIFTI_GZ'),
            inputs={
                'in_file': ('ute1_registered', nifti_gz_format)},
            requirements=[fsl_req.v('5.0.10')],
            wall_time=3)

        region_template_mask = pipeline.add(
            'region_template_mask',
            FLIRT(
                apply_xfm=True,
                bgvalue=1,
                interp='nearestneighbour',
                output_type='NIFTI_GZ'),
            inputs={
                'in_file': (zero_template_mask, 'out_file'),
                'reference': (echo1_conv, 'out_file'),
                'in_matrix_file': ('template_to_ute_mat', text_matrix_format)},
            requirements=[fsl_req.v('5.0.10')],
            wall_time=5)

        fill_in_umap = pipeline.add(
            'fill_in_umap',
            MultiImageMaths(
                op_string="-mul %s ",
                output_type='NIFTI_GZ'),
            inputs={
                'in_file': (region_template_mask, 'out_file'),
                'operand_files': (umap_conv, 'out_file')},
            requirements=[fsl_req.v('5.0.10')],
            wall_time=3)

        sute_fix_ute_space = pipeline.add(
            'sute_fix_ute_space',
            FLIRT(
                apply_xfm=True,
                bgvalue=0,
                output_type='NIFTI_GZ'),
            inputs={
                'reference': (echo1_conv, 'out_file'),
                'in_matrix_file': ('template_to_ute_mat', nifti_gz_format),
                'in_file': ('sute_fix_template', nifti_gz_format)},
            requirements=[fsl_req.v('5.0.10')],
            wall_time=5)

        sute_cont_ute_space = pipeline.add(
            'sute_cont_ute_space',
            FLIRT(
                apply_xfm=True,
                bgvalue=0,
                output_type='NIFTI_GZ'),
            inputs={
                'in_matrix_file': ('template_to_ute_mat', nifti_gz_format),
                'in_file': ('sute_cont_template', nifti_gz_format),
                'reference': (echo1_conv, 'out_file')},
            requirements=[fsl_req.v('5.0.10')],
            wall_time=5)

        sute_fix_ute_background = pipeline.add(
            'sute_fix_ute_background',
            MultiImageMaths(
                op_string="-add %s ",
                output_type='NIFTI_GZ'),
            inputs={
                'in_file': (sute_fix_ute_space, 'out_file'),
                'operand_files': (fill_in_umap, 'out_file')},
            requirements=[fsl_req.v('5.0.10')],
            wall_time=5)

        sute_cont_ute_background = pipeline.add(
            'sute_cont_ute_background',
            MultiImageMaths(
                op_string="-add %s ",
                output_type='NIFTI_GZ'),
            inputs={
                'in_file': (sute_cont_ute_space, 'out_file'),
                'operand_files': (fill_in_umap, 'out_file')},
            requirements=[fsl_req.v('5.0.10')],
            wall_time=5)

        pipeline.add(
            'smooth_sute_fix',
            Smooth(
                sigma=2.),
            inputs={
                'in_file': (sute_fix_ute_background, 'out_file')},
            requirements=[fsl_req.v('5.0.10')],
            wall_time=5)

        pipeline.add(
            'smooth_sute_cont',
            Smooth(
                sigma=2.),
            inputs={
                'in_file': (sute_cont_ute_background, 'out_file')},
            outputs={
                'sute_fix_ute': ('smoothed_file', nifti_gz_format),
                'sute_cont_ute': ('smoothed_file', nifti_gz_format)},
            requirements=[fsl_req.v('5.0.10')],
            wall_time=5)

        return pipeline



from nipype.interfaces.matlab import MatlabCommand
import banana.interfaces
from nipype.interfaces.base import (
    TraitedSpec, traits, BaseInterface, BaseInterfaceInputSpec, File,
    Directory)
import os
import os.path as op


SCRIPT_TEMPLATE = (
    "set_param(0,'CharacterEncoding','UTF-8');\n"
    "addpath(genpath('{}'));\n"
    "{{cmd}};\n"
    "exit;\n").format(op.join(banana.interfaces.RESOURCES_DIR, 'matlab',
                              'qsm'))


def matlab_cmd(cmd):
    return MatlabCommand(script=SCRIPT_TEMPLATE.format(cmd=cmd), mfile=True)


class ShMRFInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True)
    mask_file = File(exists=True, mandatory=True)


class ShMRFOutputSpec(TraitedSpec):
    out_file = File(exists=True)


class ShMRF(BaseInterface):
    input_spec = ShMRFInputSpec
    output_spec = ShMRFOutputSpec

    def _run_interface(self, runtime):  # @UnusedVariable
        mlab = matlab_cmd(
            "ShMRF('{in_file}', '{mask_file}', '{out_file}')").format(
                in_file=self.inputs.in_file,
                mask_file=self.inputs.mask_file,
                out_file=self._gen_filename('out_file'))
        result = mlab.run()
        return result.runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = self._gen_filename('out_file')

        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            fname = 'ShMRF_Vein_Mask.nii.gz'
        else:
            assert False
        return op.abspath(fname)


class FlipSWIInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True)
    hdr_file = File(exists=True, mandatory=True)


class FlipSWIOutputSpec(TraitedSpec):
    out_file = File(exists=True)


class FlipSWI(BaseInterface):
    input_spec = FlipSWIInputSpec
    output_spec = FlipSWIOutputSpec

    def _run_interface(self, runtime):  # @UnusedVariable
        mlab = matlab_cmd(
            "FlipSWI('{in_file}', '{hdr_file}', '{out_file}')".format(
                in_file=self.inputs.in_file,
                hdr_file=self.inputs.hdr_file,
                out_file=self._gen_filename('out_file')))
        result = mlab.run()
        return result.runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = self._gen_filename('out_file')

        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            fname = 'Flipped_Scanner_Image.nii.gz'
        else:
            assert False
        return op.abspath(fname)


class CVImageInputSpec(BaseInterfaceInputSpec):
    qsm = File(exists=True, mandatory=True)
    swi = File(exists=True, mandatory=True)
    mask = File(exists=True, mandatory=True)
    vein_atlas = File(exists=True, mandatory=True)
    q_prior = File(exists=True, mandatory=True)
    s_prior = File(exists=True, mandatory=True)
    a_prior = File(exists=True, mandatory=True)


class CVImageOutputSpec(TraitedSpec):
    out_file = File(exists=True)


class CVImage(BaseInterface):
    input_spec = CVImageInputSpec
    output_spec = CVImageOutputSpec

    def _run_interface(self, runtime):  # @UnusedVariable
        mlab = matlab_cmd(
            "CVImage('{qsm_file}', '{swi_file}', '{vein_atlas_file}', "
            "'{mask_file}', '{q_prior_file}', '{s_prior_file}', "
            "'{a_prior_file}', '{out_file}')".format(
                qsm_file=self.inputs.qsm,
                swi_file=self.inputs.swi,
                vein_atlas_file=self.inputs.vein_atlas,
                mask_file=self.inputs.mask,
                q_prior_file=self.inputs.q_prior,
                s_prior_file=self.inputs.s_prior,
                a_prior_file=self.inputs.a_prior,
                out_file=self._gen_filename('out_file')))
        result = mlab.run()
        return result.runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = self._gen_filename('out_file')
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            fname = 'CVImage.nii.gz'
        else:
            assert False
        return op.abspath(fname)


class PrepareInputSpec(BaseInterfaceInputSpec):
    in_dir = Directory(exists=True, mandatory=True)
    base_filename = traits.Str(
        value='T2swi3d_ axial_p2_0.9_iso_COSMOS_Straight_Coil',
        mandatory=True, desc='Base filename of coil files')
    echo_times = traits.List(traits.Float(), mandatory=True, value=[20.0],
                             desc='Echo times in ms')
    num_channels = traits.Int(value=32, mandatory=True,
                              desc='Number of channels')


class PrepareOutputSpec(TraitedSpec):
    out_dir = Directory(exists=True)
    out_file_fe = File(exists=True)
    out_file_le = File(exists=True)


class Prepare(BaseInterface):
    input_spec = PrepareInputSpec
    output_spec = PrepareOutputSpec

    def _run_interface(self, runtime):  # @UnusedVariable
        mlab = matlab_cmd(
            "Prepare_Raw_Channels('{in_dir}', '{filename}', {echo_times}, "
            "{num_channels}, '{out_dir}', '{out_file_fe}', '{out_file_le}')"
            .format(
                in_dir=self.inputs.in_dir,
                filename=self.inputs.base_filename,
                out_dir=self._gen_filename('out_dir'),
                out_file_fe=self._gen_filename('out_file_fe'),
                out_file_le=self._gen_filename('out_file_le'),
                echo_times=self.inputs.echo_times,
                num_channels=self.inputs.num_channels))
        result = mlab.run()
        return result.runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_dir'] = self._gen_filename('out_dir')
        outputs['out_file_fe'] = self._gen_filename('out_file_fe')
        outputs['out_file_le'] = self._gen_filename('out_file_le')

        return outputs

    def _gen_filename(self, name):
        if name == 'out_file_fe':
            fname = op.join('Raw', 'Raw_MAGNITUDE_FirstEcho.nii.gz')
        elif name == 'out_file_le':
            fname = op.join('Raw', 'Raw_MAGNITUDE_LastEcho.nii.gz')
        elif name == 'out_dir':
            fname = 'Raw'
        else:
            assert False
        return op.abspath(fname)


class FillHolesInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True)


class FillHolesOutputSpec(TraitedSpec):
    out_file = File(desc='Filled mask file')


class FillHoles(BaseInterface):
    input_spec = FillHolesInputSpec
    output_spec = FillHolesOutputSpec

    def _run_interface(self, runtime):  # @UnusedVariable
        mlab = matlab_cmd(
            "fillholes('{in_file}', '{out_file}');\n"
            "exit;\n").format(
                in_file=self.inputs.in_file,
                out_file=self._gen_filename('out_file'))
        result = mlab.run()
        return result.runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = self._gen_filename('out_file')
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            fname = 'Filled_Mask.nii.gz'
        else:
            assert False
        return op.abspath(fname)


class FitMaskInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True)
    initial_mask_file = File(exists=True, mandatory=True)


class FitMaskOutputSpec(TraitedSpec):
    out_file = File(desc='Fitted mask file')


class FitMask(BaseInterface):
    input_spec = FitMaskInputSpec
    output_spec = FitMaskOutputSpec

    def _run_interface(self, runtime):  # @UnusedVariable
        mlab = matlab_cmd(
            "FitMask('{in_file}', '{initial_mask_file}', '{out_file}');\n"
            "exit;\n").format(
                in_file=self.inputs.in_file,
                initial_mask_file=self.inputs.initial_mask_file,
                out_file=op.join(os.getcwd(),
                                         self._gen_filename('out_file')))
        result = mlab.run()
        return result.runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = self._gen_filename('out_file')
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            fname = 'Fitted_Mask.nii.gz'
        else:
            assert False
        return op.abspath(fname)


class QSMSummaryInputsSpec(BaseInterfaceInputSpec):
    in_field_names = traits.List(traits.Str())
    in_field_values = traits.List(traits.List(traits.List(traits.Any())))
    in_visit_id = traits.List(traits.List(traits.Str()))
    in_subject_id = traits.List(traits.List(traits.Str()))


class QSMSummaryOutputSpec(TraitedSpec):
    out_file = File(exists=True)


class QSMSummary(BaseInterface):
    input_spec = QSMSummaryInputsSpec
    output_spec = QSMSummaryOutputSpec

    def _run_interface(self, runtime):  # @UnusedVariable
        with open(op.join(os.getcwd(),
                               self._gen_filename('out_file')), 'w') as fp:

            fp.write('subjectId,visitId,' + ','.join(
                str(t) for t in self.inputs.in_field_names) + '\n')

            for s, v, o in zip(self.inputs.in_subject_id,
                               self.inputs.in_visit_id,
                               self.inputs.in_field_values):
                for ts, tv, to in zip(s, v, o):
                    fp.write(','.join(str(t) for t in [ts, tv]) + ',')
                    fp.write(','.join(str(t) for t in to) + '\n')
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = self._gen_filename('out_file')
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            fname = 'qsm_summary.csv'
        else:
            assert False
        return op.abspath(fname)



import os
import os.path as op
import logging
import tempfile
from copy import copy
import json
from pprint import pformat
from itertools import chain
from unittest import TestCase
from argparse import ArgumentParser
from importlib import import_module
from arcana.exceptions import ArcanaNameError
from arcana import (FilesetInput, FieldInput, BasicRepo, XnatRepo, SingleProc,
                    Field, Fileset)
from arcana.exceptions import ArcanaInputMissingMatchError
from banana.exceptions import BananaTestSetupError, BananaUsageError
import banana.file_format  # @UnusedImport


logger = logging.getLogger('arcana')
logger.setLevel(logging.WARNING)
handler = logging.StreamHandler()
formatter = logging.Formatter("%(levelname)s - %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)

logging.getLogger("urllib3").setLevel(logging.WARNING)

try:
    TEST_DIR = os.environ['BANANA_TEST_DIR']
except KeyError:
    TEST_DIR = tempfile.mkdtemp()

TEST_CACHE_DIR = op.join(TEST_DIR, 'cache')


class PipelineTester(TestCase):
    """
    Runs pipelines within a Study class and compares the results
    against reference data from previous runs

    Class attributes
    ----------------
    name : str
        A unique name (among other unittests to be run in same batch) for the
        test class
    study_class : type(Study)
        The study class run tests against
    ref_repo : Repository
        The repository to draw the reference data from
    parameters : dict[str, str | float | int | datetime]
        The parameters passed to the study when it is initialised
    warn_missing_tests : bool
        Whether to raise a warning if there is a pipeline in the study class
        that does not have a test in tester class
    """

    parameters = {}
    warn_missing_tests = True

    def setUp(self):
        if not hasattr(self, 'study_class'):
            raise BananaTestSetupError(
                "{} must have a 'study_class' attribute corresponding to the "
                "Study class to test".format(type(self).__name__))
        if not hasattr(self, 'ref_repo'):
            raise BananaTestSetupError(
                "{} must have a 'ref_repo' attribute corresponding to the "
                "Study class to test".format(type(self).__name__))
        base_test_dir = op.join(TEST_DIR, self.name)
        self.work_dir = op.join(base_test_dir, 'work')
        repo_root = op.join(base_test_dir, 'repo')
        os.makedirs(self.work_dir, exist_ok=True)
        os.makedirs(repo_root, exist_ok=True)
        # Create repository to hold outputs
        self.output_repo = BasicRepo(repo_root, depth=2)
        # Create inputs for reference study
        self.inputs = {}
        for spec in self.study_class.data_specs():
            # Create an input for each entry in the class specificiation
            if spec.is_fileset:
                selector = FilesetInput(
                    spec.name, spec.name, repository=self.ref_repo)
            else:
                selector = FieldInput(
                    spec.name, spec.name, dtype=spec.dtype,
                    repository=self.ref_repo)
            # Check whether a corresponding data exists in the reference repo
            try:
                selector.match(self.ref_repo.cached_tree(), spec)
            except ArcanaInputMissingMatchError:
                continue
            self.inputs[spec.name] = selector
        # Create the reference study
        self.ref_study = self.study_class(
            self.name,
            repository=self.ref_repo,
            processor=self.work_dir,
            inputs=self.inputs.values(),
            parameters=self.parameters)
        # Get set of all pipelines to test
        if self.warn_missing_tests:
            pipelines_to_test = set(
                s.pipeline_getter for s in self.study_class.data_specs()
                if s.derived)
            test_names = [m for m in dir(self) if m.startswith('test_')]
            for pipeline_getter in pipelines_to_test:
                expected_test_name = 'test_' + pipeline_getter
                if expected_test_name not in test_names:
                    logger.warning("Did not find test for '{}' pipeline in {}"
                                   "tester. Add to 'suppress_warning' class "
                                   "attr to suppress this message".format(
                                       pipeline_getter, self.name))

    def run_pipeline_test(self, pipeline_getter, add_inputs=(),
                          test_criteria=None):
        """
        Runs a pipeline and tests its outputs against the reference data

        Parameters
        ----------
        pipeline_getter : str
            The name of the pipeline to test
        add_inputs : list[str]
            Inputs that are required in the output study for the pipeline to
            run, in addition to the direct inputs of the pipeline, i.e. ones
            that are tested for with the 'provided' method in the pipeline
            construction
        test_criteria : dct[str, *] | None
            A dictionary containing the criteria by which to determine equality
            for the derived filesets. The keys are spec-names and the values
            are specific to the format of the fileset. If a spec-name is not
            in the dictionary or None is provided then the default
            criteria are used for each fileset test.
        """
        if test_criteria is None:
            test_criteria = {}
        # A study with all inputs provided to determine which inputs are needed
        # by the pipeline
        ref_pipeline = self.ref_study.pipeline(pipeline_getter)
        inputs = []
        for spec_name in chain(ref_pipeline.input_names, add_inputs):
            try:
                inputs.append(self.inputs[spec_name])
            except KeyError:
                pass  # Inputs with a default value
        # Set up output study
        output_study = self.study_class(
            pipeline_getter,
            repository=self.output_repo,
            processor=SingleProc(self.work_dir, reprocess='force'),
            inputs=inputs,
            parameters=self.parameters,
            subject_ids=self.ref_study.subject_ids,
            visit_ids=self.ref_study.visit_ids,
            enforce_inputs=False,
            fill_tree=True)
        for spec_name in ref_pipeline.output_names:
            for ref, test in zip(self.ref_study.data(spec_name),
                                 output_study.data(spec_name)):
                if ref.is_fileset:
                    try:
                        self.assertTrue(
                            test.contents_equal(
                                ref, **test_criteria.get(spec_name, {})),
                            "'{}' fileset generated by {} in {} doesn't match "
                            "reference".format(spec_name, pipeline_getter,
                                               self.study_class))
                    except Exception:
                        if hasattr(test, 'headers_diff'):
                            header_diff = test.headers_diff(ref)
                            if header_diff:
                                print("Headers don't match on {}"
                                      .format(header_diff))
                                print("Test header:\n{}".format(
                                    pformat(test.get_header())))
                                print("Reference header:\n{}".format(
                                    pformat(ref.get_header())))
                            else:
                                print("Image RMS diff: {}"
                                      .format(test.rms_diff(ref)))
                                test.contents_equal(ref)
                        raise
                else:
                    self.assertEqual(
                        test.value, ref.value,
                        "value for {} ({}) generated by {} in {} doesn't "
                        "match reference ({})".format(
                            spec_name, test.value, pipeline_getter,
                            self.study_class, ref.value))

    @classmethod
    def generate_test_data(cls, study_class, in_repo, out_repo,
                           in_server=None, out_server=None, work_dir=None,
                           parameters=(), include=None, skip=(),
                           reprocess=False, repo_depth=0):
        """
        Generates reference data for a pipeline tester unittests given a study
        class and set of parameters

        Parameters
        ----------
        study_class : type(Study)
            The path to the study class to test, e.g. banana.study.MriStudy
        in_repo : str
            The path to repository that houses the input data
        out_repo : str
            If the 'xnat_server' argument is provided then out
            is interpreted as the project ID to use the XNAT
            server (the project must exist already). Otherwise
            it is interpreted as the path to a basic repository
        in_server : str | None
            The server to download the input data from
        out_server : str | None
            The server to upload the reference data to
        work_dir : str
            The work directory
        parameters : dict[str, *]
            Parameter to set when initialising the study
        include : list[str] | None
            Spec names to include in the output repository. If None all names
            except those listed in 'skip' are included
        skip : list[str]
            Spec names to skip in the generation process. Only valid if
            'include' is None
        reprocess : bool
            Whether to reprocess the generated datasets
        repo_depth : int
            The depth of the input repository
        """
        if work_dir is None:
            work_dir = tempfile.mkdtemp()
        else:
            work_dir = work_dir

        parts = study_class.split('.')
        module_name = '.'.join(parts[:-1])
        class_name = parts[-1]

        # Convert parameters to dictionary
        parameters_dct = {}
        for name, value in parameters:
            try:
                value = int(value)
            except ValueError:
                try:
                    value = float(value)
                except ValueError:
                    pass
            parameters_dct[name] = value
        parameters = parameters_dct

        module = import_module(module_name)
        study_class = getattr(module, class_name)

        if class_name.endswith('Study'):
            study_name = class_name[:-len('Study')]
        else:
            study_name = class_name

        # Get output repository to write the data to
        if in_server is not None:
            in_repo = XnatRepo(project_id=in_repo, server=in_server,
                               cache_dir=op.join(work_dir, 'xnat-cache'))
        else:
            in_repo = BasicRepo(in_repo, depth=repo_depth)

        temp_repo = BasicRepo(op.join(work_dir, 'temp-repo'), depth=repo_depth)

        in_paths = []
        inputs = []
        for fname in os.listdir(in_repo):
            if fname == study_name:
                continue
            path = op.join(in_repo.root_dir, fname)
            in_paths.append(path)
            if fname == BasicRepo.FIELDS_FNAME:
                with open(path) as f:
                    field_data = json.load(f)
                for name, value in field_data.items():
                    field = Field(name, value)
                    selector = FieldInput(field.name, field.name, field.dtype,
                                          repository=in_repo)
                    inputs.append(selector)
            else:
                fileset = Fileset.from_path(path)
                selector = FilesetInput(fileset.basename, fileset.basename,
                                        fileset.format, repository=in_repo)
                try:
                    spec = study_class.data_spec(selector)
                except ArcanaNameError:
                    print("Skipping '{}' as it doesn't match a spec in {}"
                          .format(fname, study_class))
                    continue
                if not (spec.derived and
                        selector.format in spec.valid_formats):
                    print("Skipping '{}' as it doesn't have a valid format "
                          "for {} in {} ({})".format(
                              fname, spec.name, study_class,
                              ', '.join(f.name for f in spec.valid_formats)))
                    continue
                inputs.append(selector)

        study = study_class(
            study_name,
            repository=temp_repo,
            processor=SingleProc(work_dir, reprocess=reprocess),
            inputs=inputs,
            parameters=parameters,
            fill_tree=True)

        if include is None:
            include = [n for n in study.data_spec_names() if n not in skip]
        elif skip:
            raise BananaUsageError(
                "Cannot provide both 'include' and 'skip' options to "
                "PipelineTester.generate_test_data")

        # Generate all derived data
        for spec_name in include:
            study.data(spec_name)

        # Get output repository to write the data to
        if out_server is not None:
            out_repo = XnatRepo(project_id=out_repo, server=in_server,
                                cache_dir=op.join(work_dir, 'xnat-cache'))
        else:
            out_repo = BasicRepo(out_repo, depth=repo_depth)

        # Upload data to repository
        for spec in study.data_specs():
            if spec.name not in skip:
                for item in study.data(spec.name):
                    if not item.exists:
                        continue
                    if item.is_fileset:
                        item_cpy = Fileset(
                            name=item.name, format=item.format,
                            frequency=item.frequency, path=item.path,
                            aux_files=copy(item.aux_files),
                            subject_id=item.subject_id, visit_id=item.visit_id,
                            repository=out_repo, exists=True)
                    else:
                        item_cpy = Field(
                            name=item.name, value=item.value, dtype=item.dtype,
                            frequency=item.frequency, array=item.array,
                            subject_id=item.subject_id, visit_id=item.visit_id,
                            repository=out_repo, exists=True)
                    item_cpy.put()


def gen_test_data_entry_point():
    parser = ArgumentParser(
        description=("Generates reference data for a pipeline tester "
                     "unittests given a study class and set of parameters"))
    parser.add_argument('study_class',
                        help=("The path to the study class to test, e.g. "
                              "banana.study.MriStudy"))
    parser.add_argument('in_repo', help=("The path to repository that "
                                         "houses the input data"))
    parser.add_argument('out_repo',
                        help=("If the 'xnat_server' argument is provided then "
                              "out is interpreted as the project ID to use "
                              "the XNAT server (the project must exist "
                              "already). Otherwise it is interpreted as the "
                              "path to a basic repository"))
    parser.add_argument('--xnat_server', default=None,
                        help="The server to upload the reference data to")
    parser.add_argument('--work_dir', default=None,
                        help="The work directory")
    parser.add_argument('--parameter', '-p', metavar=('NAME', 'VALUE'),
                        nargs=2, action='append', default=[],
                        help=("Parameters to set when initialising the "
                              "study"))
    parser.add_argument('--skip', '-s', nargs='+', default=[],
                        help=("Spec names to skip in the generation "
                              "process"))
    parser.add_argument('--reprocess', action='store_true', default=False,
                        help=("Whether to reprocess previously generated "
                              "datasets in the output repository"))
    parser.add_argument('--repo_depth', type=int, default=0,
                        help="The depth of the input repository")
    args = parser.parse_args()

    PipelineTester.generate_test_data(
        study_class=args.study_class, in_repo=args.in_repo,
        out_repo=args.out_repo, xnat_server=args.xnat_server,
        work_dir=args.work_dir, parameters=args.parameter, skip=args.skip,
        reprocess=args.reprocess, repo_depth=args.repo_depth)
